# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import os

import jsonschema

from rally.common import logging
from rally.common import validation

LOG = logging.getLogger(__name__)


@validation.configure(name="jsonschema")
class JsonSchemaValidator(validation.Validator):
    """JSON schema validator"""

    def validate(self, context, config, plugin_cls, plugin_cfg):
        try:
            jsonschema.validate(plugin_cfg, plugin_cls.CONFIG_SCHEMA)
        except jsonschema.ValidationError as err:
            self.fail(str(err))


@validation.configure(name="args-spec")
class ArgsValidator(validation.Validator):
    """Scenario arguments validator"""

    def validate(self, context, config, plugin_cls, plugin_cfg):
        scenario = plugin_cls
        name = scenario.get_name()
        platform = scenario.get_platform()

        args_spec = inspect.signature(scenario().run).parameters
        missed_args = [
            p.name for p in args_spec.values()
            if p.default == inspect.Parameter.empty
            and p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD]

        hint_msg = (" Use `rally plugin show --name %s --platform %s` "
                    "to display scenario description." % (name, platform))

        if "args" in config:
            missed_args = set(missed_args) - set(config["args"])
        if missed_args:
            msg = ("Argument(s) '%(args)s' should be specified in task config."
                   "%(hint)s" % {"args": "', '".join(missed_args),
                                 "hint": hint_msg})
            self.fail(msg)

        support_kwargs = any(
            p for p in args_spec.values()
            if p.kind == inspect.Parameter.VAR_KEYWORD
        )

        if not support_kwargs and "args" in config:
            redundant_args = [p for p in config["args"] if p not in args_spec]
            if redundant_args:
                msg = ("Unexpected argument(s) found ['%(args)s'].%(hint)s" %
                       {"args": "', '".join(redundant_args),
                        "hint": hint_msg})
                self.fail(msg)


@validation.configure(name="required_params")
class RequiredParameterValidator(validation.Validator):
    """Scenario required parameter validator.

    This allows us to search required parameters in subdict of config.

    :param subdict: sub-dict of "config" to search. if
                    not defined - will search in "config"
    :param params: list of required parameters
    """

    def __init__(self, params=None, subdict=None):
        super(RequiredParameterValidator, self).__init__()
        self.subdict = subdict
        self.params = params

    def validate(self, context, config, plugin_cls, plugin_cfg):
        missing = []
        args = config.get("args", {})
        if self.subdict:
            args = args.get(self.subdict, {})
        for arg in self.params:
            if isinstance(arg, (tuple, list)):
                for case in arg:
                    if case in args:
                        break
                else:
                    arg = "'/'".join(arg)
                    missing.append("'%s' (at least one parameter should be "
                                   "specified)" % arg)
            else:
                if arg not in args:
                    missing.append("'%s'" % arg)

        if missing:
            msg = ("%s parameter(s) are not defined in "
                   "the input task file") % ", ".join(missing)
            self.fail(msg)


@validation.configure(name="number")
class NumberValidator(validation.Validator):
    """Checks that parameter is a number that pass specified condition.

    Ensure a parameter is within the range [minval, maxval]. This is a
    closed interval so the end points are included.

    :param param_name: Name of parameter to validate
    :param minval: Lower endpoint of valid interval
    :param maxval: Upper endpoint of valid interval
    :param nullable: Allow parameter not specified, or parameter=None
    :param integer_only: Only accept integers
    """

    def __init__(self, param_name, minval=None, maxval=None, nullable=False,
                 integer_only=False):
        self.param_name = param_name
        self.minval = minval
        self.maxval = maxval
        self.nullable = nullable
        self.integer_only = integer_only

    def validate(self, context, config, plugin_cls, plugin_cfg):

        value = config.get("args", {}).get(self.param_name)

        num_func = float
        if self.integer_only:
            # NOTE(boris-42): Force check that passed value is not float, this
            #   is important cause int(float_numb) won't raise exception
            if type(value) == float:
                return self.fail("%(name)s is %(val)s which hasn't int type"
                                 % {"name": self.param_name, "val": value})
            num_func = int

        # None may be valid if the scenario sets a sensible default.
        if self.nullable and value is None:
            return

        try:
            number = num_func(value)
            if self.minval is not None and number < self.minval:
                self.fail("%(name)s is %(val)s which is less than the minimum "
                          "(%(min)s)" % {"name": self.param_name,
                                         "val": number,
                                         "min": self.minval})
            if self.maxval is not None and number > self.maxval:
                self.fail("%(name)s is %(val)s which is greater than the "
                          "maximum (%(max)s)" % {"name": self.param_name,
                                                 "val": number,
                                                 "max": self.maxval})
        except (ValueError, TypeError):
            self.fail("%(name)s is %(val)s which is not a valid %(type)s" %
                      {"name": self.param_name, "val": value,
                       "type": num_func.__name__})


@validation.configure(name="enum")
class EnumValidator(validation.Validator):
    """Checks that parameter is in a list.

    Ensure a parameter has the right value. This value need to be defined
    in a list.

    :param param_name: Name of parameter to validate
    :param values: List of values accepted
    :param missed: Allow to accept optional parameter
    :param case_insensitive: Ignore case in enum values
    """

    def __init__(self, param_name, values, missed=False,
                 case_insensitive=False):
        self.param_name = param_name
        self.missed = missed
        self.case_insensitive = case_insensitive
        if self.case_insensitive:
            self.values = []
            for value in values:
                if isinstance(value, str):
                    value = value.lower()
                self.values.append(value)
        else:
            self.values = values

    def validate(self, context, config, plugin_cls, plugin_cfg):
        value = config.get("args", {}).get(self.param_name)
        if value:
            if self.case_insensitive:
                if isinstance(value, str):
                    value = value.lower()

            if value not in self.values:
                self.fail("%(name)s is %(val)s which is not a valid value "
                          "from %(list)s" % {"name": self.param_name,
                                             "val": value,
                                             "list": self.values})
        else:
            if not self.missed:
                self.fail("%s parameter is not defined in the task config file"
                          % self.param_name)


@validation.configure(name="map_keys")
class MapKeysParameterValidator(validation.Validator):
    """Check that parameter contains specified keys.

    :param param_name: Name of parameter to validate
    :param required: List of all required keys
    :param allowed: List of all allowed keys
    :param additional: Whether additional keys are allowed. If list of allowed
           keys are specified, defaults to False, otherwise defaults to True
    :param missed: Allow to accept optional parameter
    """
    def __init__(self, param_name, required=None, allowed=None,
                 additional=True, missed=False):
        super(MapKeysParameterValidator, self).__init__()
        self.param_name = param_name
        self.required = required or []
        self.allowed = allowed or []
        self.additional = additional
        self.missed = missed

    def validate(self, context, config, plugin_cls, plugin_cfg):
        parameter = config.get("args", {}).get(self.param_name)

        if parameter:
            required_diff = set(self.required) - set(parameter.keys())
            if required_diff:
                self.fail(
                    "Required keys is missing in '%(name)s' parameter: "
                    "%(key)s" % {"name": self.param_name,
                                 "key": ", ".join(sorted(list(required_diff)))}
                )

            if self.allowed:
                allowed_diff = set(parameter.keys()) - set(self.allowed)
                if allowed_diff:
                    self.fail(
                        "Parameter '%(name)s' contains unallowed keys: "
                        "%(key)s" % {
                            "name": self.param_name,
                            "key": ", ".join(sorted(list(allowed_diff)))}
                    )
            elif not self.additional:
                diff = set(parameter.keys()) - set(self.required)
                if diff:
                    self.fail(
                        "Parameter '%(name)s' contains unallowed keys: "
                        "%(key)s" % {
                            "name": self.param_name,
                            "key": ", ".join(sorted(list(diff)))}
                    )
        elif not self.missed:
            self.fail("'%s' parameter is not defined in the task config file"
                      % self.param_name)


@validation.configure(name="restricted_parameters")
class RestrictedParametersValidator(validation.Validator):

    def __init__(self, param_names, subdict=None):
        """Validates that parameters is not set.

        :param param_names: parameter or parameters list to be validated.
        :param subdict: sub-dict of "config" to search for param_names. if
                        not defined - will search in "config"
        """
        super(RestrictedParametersValidator, self).__init__()
        if isinstance(param_names, (list, tuple)):
            self.params = param_names
        else:
            self.params = [param_names]
        self.subdict = subdict

    def validate(self, context, config, plugin_cls, plugin_cfg):
        restricted_params = []
        for param_name in self.params:
            args = config.get("args", {})
            a_dict, a_key = (args, self.subdict) if self.subdict else (
                config, "args")
            if param_name in a_dict.get(a_key, {}):
                restricted_params.append(param_name)
        if restricted_params:
            self.fail("You can't specify parameters '%s' in '%s'" % (
                ", ".join(restricted_params),
                self.subdict if self.subdict else "args"))


@validation.configure(name="required_contexts")
class RequiredContextsValidator(validation.Validator):

    def __init__(self, contexts, *args):
        """Validator checks if required contexts are specified.

        :param contexts: list of strings and tuples with context names that
                         should be specified. Tuple represent 'at least one
                         of the'.
        """
        super(RequiredContextsValidator, self).__init__()
        if isinstance(contexts, (list, tuple)):
            # services argument is a list, so it is a new way of validators
            #  usage, args in this case should not be provided
            self.contexts = contexts
            if args:
                LOG.warning("Positional argument is not what "
                            "'required_context' decorator expects. "
                            "Use `contexts` argument instead")
        else:
            # it is old way validator
            self.contexts = [contexts]
            self.contexts.extend(args)

    def validate(self, context, config, plugin_cls, plugin_cfg):
        missing_contexts = []
        input_context = config.get("contexts", {})

        for name in self.contexts:
            if isinstance(name, tuple):
                if not set(name) & set(input_context):
                    # formatted string like: 'foo or bar or baz'
                    formatted_names = "'%s'" % " or ".join(name)
                    missing_contexts.append(formatted_names)
            else:
                if name not in input_context:
                    missing_contexts.append(name)

        if missing_contexts:
            self.fail("The following context(s) are required but missing from "
                      "the input task file: %s" % ", ".join(missing_contexts))


@validation.configure(name="required_param_or_context")
class RequiredParamOrContextValidator(validation.Validator):

    def __init__(self, param_name, ctx_name):
        """Validator checks if required image is specified.

        :param param_name: name of parameter
        :param ctx_name: name of context
        """
        super(RequiredParamOrContextValidator, self).__init__()
        self.param_name = param_name
        self.ctx_name = ctx_name

    def validate(self, context, config, plugin_cls, plugin_cfg):
        msg = ("You should specify either scenario argument %s or"
               " use context %s." % (self.param_name, self.ctx_name))

        if self.ctx_name in config.get("contexts", {}):
            return
        if self.param_name in config.get("args", {}):
            return
        self.fail(msg)


@validation.configure(name="file_exists")
class FileExistsValidator(validation.Validator):

    def __init__(self, param_name, mode=os.R_OK, required=True):
        """Validator checks parameter is proper path to file with proper mode.

        Ensure a file exists and can be accessed with the specified mode.
        Note that path to file will be expanded before access checking.

        :param param_name: Name of parameter to validate
        :param mode: Access mode to test for. This should be one of:
            * os.F_OK (file exists)
            * os.R_OK (file is readable)
            * os.W_OK (file is writable)
            * os.X_OK (file is executable)

            If multiple modes are required they can be added, eg:
                mode=os.R_OK+os.W_OK
        :param required: Boolean indicating whether this argument is required.
        """
        super(FileExistsValidator, self).__init__()

        self.param_name = param_name
        self.mode = mode
        self.required = required

    def _file_access_ok(self, filename, mode, param_name, required=True):
        if not filename:
            if not required:
                return
            self.fail("Parameter %s required" % param_name)
        if not os.access(os.path.expanduser(filename), mode):
            self.fail("Could not open %(filename)s with mode %(mode)s for "
                      "parameter %(param_name)s" % {"filename": filename,
                                                    "mode": mode,
                                                    "param_name": param_name})

    def validate(self, context, config, plugin_cls, plugin_cfg):

        self._file_access_ok(config.get("args", {}).get(self.param_name),
                             self.mode, self.param_name, self.required)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.common import version
from tests.unit import test


class ModuleTestCase(test.TestCase):

    VERSION_REGEX = r"^\d+\.\d+\.\d+(~dev\d+)?$"

    def test_version_info(self):
        version_str = version.version_info.semantic_version().debian_string()
        self.assertRegex(version_str, self.VERSION_REGEX)

    @mock.patch("rally.common.version.version_info")
    def test_version_string(self, mock_version_info):
        mock_sv = mock.Mock()
        mock_sv.debian_string.return_value = "foo_version"
        mock_version_info.semantic_version.return_value = mock_sv
        self.assertEqual("foo_version", version.version_string())

    @mock.patch("rally.common.db.schema.schema_revision", return_value="foo")
    def test_database_revision(self, mock_schema_revision):
        self.assertEqual("foo", version.database_revision())
        mock_schema_revision.assert_called_once_with(detailed=True)



# Copyright 2013: ITLook, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import tempfile

import testtools

from tests.functional import utils


class EnvTestCase(testtools.TestCase):

    def test_create_no_spec(self):
        rally = utils.Rally()
        rally("env create --name empty --description de")
        self.assertIn("empty", rally("env list"))
        env_data = rally("env show --json", getjson=True)
        self.assertEqual("empty", env_data["name"])
        self.assertEqual("de", env_data["description"])
        self.assertEqual({}, env_data["extras"])
        self.assertEqual({}, env_data["platforms"])

    def _create_spec(self, spec):
        f = tempfile.NamedTemporaryFile(mode="w", delete=False)

        def unlink():
            os.unlink(f.name)

        self.addCleanup(unlink)

        f.write(json.dumps(spec, indent=2))
        f.close()
        return f.name

    def test_create_check_info_destroy_delete_with_spec(self):
        rally = utils.Rally(plugin_path="tests/functional/extra")

        spec = self._create_spec({"good@fake": {}})
        rally("env create --name real --spec %s" % spec)
        env = rally("env show --json", getjson=True)
        self.assertIn("fake", env["platforms"])

        env_info = rally("env info --json", getjson=True)
        self.assertEqual({"good@fake": {"info": {"a": 1}}}, env_info)

        rally("env check --json")

    def test_list_empty(self):
        rally = utils.Rally()
        # TODO(boris-42): Clean this up somehow
        rally("env destroy MAIN")
        rally("env delete MAIN")
        self.assertEqual([], rally("env list --json", getjson=True))
        self.assertIn("There are no environments", rally("env list"))

    def test_list(self):
        rally = utils.Rally()
        envs = rally("env list --json", getjson=True)
        self.assertEqual(1, len(envs))
        self.assertEqual("MAIN", envs[0]["name"])
        self.assertIn("MAIN", rally("env list"))

    def test_use(self):

        def show_helper():
            return rally("env show --json", getjson=True)

        rally = utils.Rally()
        self.assertEqual("MAIN", show_helper()["name"])
        empty_uuid = rally("env create --name empty --json",
                           getjson=True)["uuid"]
        self.assertEqual("empty", show_helper()["name"])
        rally("env use MAIN")
        self.assertEqual("MAIN", show_helper()["name"])
        rally("env use %s" % empty_uuid)
        self.assertEqual("empty", show_helper()["name"])
        rally("env create --name empty2 --description de --no-use")
        self.assertEqual("empty", show_helper()["name"])



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import os
import socket

import ddt
import mock
import paramiko

from rally import exceptions
from rally.utils import sshutils
from tests.unit import test


class FakeParamikoException(Exception):
    pass


class SSHTestCase(test.TestCase):
    """Test all small SSH methods."""

    def setUp(self):
        super(SSHTestCase, self).setUp()
        self.ssh = sshutils.SSH("root", "example.net")

    @mock.patch("rally.utils.sshutils.SSH._get_pkey")
    def test_construct(self, mock_ssh__get_pkey):
        mock_ssh__get_pkey.return_value = "pkey"
        ssh = sshutils.SSH("root", "example.net", port=33, pkey="key",
                           key_filename="kf", password="secret")
        mock_ssh__get_pkey.assert_called_once_with("key")
        self.assertEqual("root", ssh.user)
        self.assertEqual("example.net", ssh.host)
        self.assertEqual(33, ssh.port)
        self.assertEqual("pkey", ssh.pkey)
        self.assertEqual("kf", ssh.key_filename)
        self.assertEqual("secret", ssh.password)

    def test_construct_default(self):
        self.assertEqual("root", self.ssh.user)
        self.assertEqual("example.net", self.ssh.host)
        self.assertEqual(22, self.ssh.port)
        self.assertIsNone(self.ssh.pkey)
        self.assertIsNone(self.ssh.key_filename)
        self.assertIsNone(self.ssh.password)

    @mock.patch("rally.utils.sshutils.paramiko")
    def test__get_pkey_invalid(self, mock_paramiko):
        mock_paramiko.SSHException = FakeParamikoException
        rsa = mock_paramiko.rsakey.RSAKey
        dss = mock_paramiko.dsskey.DSSKey
        rsa.from_private_key.side_effect = mock_paramiko.SSHException
        dss.from_private_key.side_effect = mock_paramiko.SSHException
        self.assertRaises(exceptions.SSHError, self.ssh._get_pkey, "key")

    def test__get_pkey_dss(self):
        private_dss_key = io.StringIO()
        private_dss_key_obj = paramiko.DSSKey.generate(1024)
        private_dss_key_obj.write_private_key(private_dss_key)
        private_dss_key.seek(0)
        ssh = sshutils.SSH("root", "example.net")
        self.assertIsInstance(ssh._get_pkey(private_dss_key),
                              paramiko.DSSKey)

        private_dss_key.seek(0)
        self.assertIsInstance(ssh._get_pkey(private_dss_key.getvalue()),
                              paramiko.DSSKey)

    def test__get_pkey_rsa(self):
        private_rsa_key = io.StringIO()
        private_rsa_key_obj = paramiko.RSAKey.generate(1024)
        private_rsa_key_obj.write_private_key(private_rsa_key)
        private_rsa_key.seek(0)
        ssh = sshutils.SSH("root", "example.net")
        self.assertIsInstance(ssh._get_pkey(private_rsa_key),
                              paramiko.RSAKey)

        private_rsa_key.seek(0)
        self.assertIsInstance(ssh._get_pkey(private_rsa_key.getvalue()),
                              paramiko.RSAKey)

    @mock.patch("rally.utils.sshutils.SSH._get_pkey")
    @mock.patch("rally.utils.sshutils.paramiko")
    def test__get_client(self, mock_paramiko, mock_ssh__get_pkey):
        mock_ssh__get_pkey.return_value = "key"
        fake_client = mock.Mock()
        mock_paramiko.SSHClient.return_value = fake_client
        mock_paramiko.AutoAddPolicy.return_value = "autoadd"

        ssh = sshutils.SSH("admin", "example.net", pkey="key")
        client = ssh._get_client()

        self.assertEqual(fake_client, client)
        client_calls = [
            mock.call.set_missing_host_key_policy("autoadd"),
            mock.call.connect("example.net", username="admin",
                              port=22, pkey="key", key_filename=None,
                              password=None, timeout=1),
        ]
        self.assertEqual(client_calls, client.mock_calls)

    def test_close(self):
        with mock.patch.object(self.ssh, "_client") as m_client:
            self.ssh.close()
        m_client.close.assert_called_once_with()
        self.assertFalse(self.ssh._client)

    @mock.patch("rally.utils.sshutils.io.StringIO")
    def test_execute(self, mock_string_io):
        mock_string_io.side_effect = stdio = [mock.Mock(), mock.Mock()]
        stdio[0].read.return_value = "stdout fake data"
        stdio[1].read.return_value = "stderr fake data"
        with mock.patch.object(self.ssh, "run") as mock_run:
            mock_run.return_value = (0, None)
            status, stdout, stderr = self.ssh.execute("cmd",
                                                      stdin="fake_stdin",
                                                      timeout=43)
        mock_run.assert_called_once_with(
            "cmd", stdin="fake_stdin", stdout=stdio[0],
            stderr=stdio[1], timeout=43, raise_on_error=False)
        self.assertEqual(0, status)
        self.assertEqual("stdout fake data", stdout)
        self.assertEqual("stderr fake data", stderr)

    @mock.patch("rally.utils.sshutils.time")
    def test_wait_timeout(self, mock_time):
        mock_time.time.side_effect = [1, 50, 150]
        self.ssh.execute = mock.Mock(side_effect=[exceptions.SSHError,
                                                  exceptions.SSHError,
                                                  0])
        self.assertRaises(exceptions.SSHTimeout, self.ssh.wait)
        self.assertEqual([mock.call("uname")] * 2, self.ssh.execute.mock_calls)

    @mock.patch("rally.utils.sshutils.time")
    def test_wait(self, mock_time):
        mock_time.time.side_effect = [1, 50, 100]
        self.ssh.execute = mock.Mock(side_effect=[exceptions.SSHError,
                                                  exceptions.SSHError,
                                                  0])
        self.ssh.wait()
        self.assertEqual([mock.call("uname")] * 3, self.ssh.execute.mock_calls)


@ddt.ddt
class SSHRunTestCase(test.TestCase):
    """Test SSH.run method in different aspects.

    Also tested method "execute".
    """

    def setUp(self):
        super(SSHRunTestCase, self).setUp()

        self.fake_client = mock.Mock()
        self.fake_session = mock.Mock()
        self.fake_transport = mock.Mock()

        self.fake_transport.open_session.return_value = self.fake_session
        self.fake_client.get_transport.return_value = self.fake_transport

        self.fake_session.recv_ready.return_value = False
        self.fake_session.recv_stderr_ready.return_value = False
        self.fake_session.send_ready.return_value = False
        self.fake_session.exit_status_ready.return_value = True
        self.fake_session.recv_exit_status.return_value = 0

        self.ssh = sshutils.SSH("admin", "example.net")
        self.ssh._get_client = mock.Mock(return_value=self.fake_client)

    @mock.patch("rally.utils.sshutils.select")
    def test_execute(self, mock_select):
        mock_select.select.return_value = ([], [], [])
        self.fake_session.recv_ready.side_effect = [1, 0, 0]
        self.fake_session.recv_stderr_ready.side_effect = [1, 0]
        self.fake_session.recv.return_value = b"ok"
        self.fake_session.recv_stderr.return_value = b"error"
        self.fake_session.exit_status_ready.return_value = 1
        self.fake_session.recv_exit_status.return_value = 127
        self.assertEqual((127, "ok", "error"), self.ssh.execute("cmd"))
        self.fake_session.exec_command.assert_called_once_with("cmd")

    @mock.patch("rally.utils.sshutils.select")
    def test_execute_args(self, mock_select):
        mock_select.select.return_value = ([], [], [])
        self.fake_session.recv_ready.side_effect = [1, 0, 0]
        self.fake_session.recv_stderr_ready.side_effect = [1, 0]
        self.fake_session.recv.return_value = b"ok"
        self.fake_session.recv_stderr.return_value = b"error"
        self.fake_session.exit_status_ready.return_value = 1
        self.fake_session.recv_exit_status.return_value = 127

        result = self.ssh.execute(["cmd", "arg1", "arg2 with space"])
        self.assertEqual((127, "ok", "error"), result)
        self.fake_session.exec_command.assert_called_once_with(
            "cmd arg1 'arg2 with space'")

    @mock.patch("rally.utils.sshutils.select")
    def test_run(self, mock_select):
        mock_select.select.return_value = ([], [], [])
        exit_stat, data = self.ssh.run("cmd")
        self.assertEqual(0, exit_stat)

    @mock.patch("rally.utils.sshutils.select")
    def test_run_nonzero_status(self, mock_select):
        mock_select.select.return_value = ([], [], [])
        self.fake_session.recv_exit_status.return_value = 1
        self.assertRaises(exceptions.SSHError, self.ssh.run, "cmd")
        exit_stat, data = self.ssh.run("cmd", raise_on_error=False)
        self.assertEqual(1, exit_stat)

    @mock.patch("rally.utils.sshutils.select")
    def test_run_stdout(self, mock_select):
        mock_select.select.return_value = ([], [], [])
        self.fake_session.recv_ready.side_effect = [True, True, False]
        self.fake_session.recv.side_effect = [b"ok1", b"ok2"]
        stdout = mock.Mock()
        self.ssh.run("cmd", stdout=stdout)
        self.assertEqual([mock.call("ok1"), mock.call("ok2")],
                         stdout.write.mock_calls)

    @mock.patch("rally.utils.sshutils.select")
    def test_run_stderr(self, mock_select):
        mock_select.select.return_value = ([], [], [])
        self.fake_session.recv_stderr_ready.side_effect = [True, False]
        self.fake_session.recv_stderr.return_value = b"error"
        stderr = mock.Mock()
        self.ssh.run("cmd", stderr=stderr)
        stderr.write.assert_called_once_with("error")

    @mock.patch("rally.utils.sshutils.select")
    def test_run_stdin(self, mock_select):
        """Test run method with stdin.

        Third send call was called with "e2" because only 3 bytes was sent
        by second call. So remainig 2 bytes of "line2" was sent by third call.
        """
        mock_select.select.return_value = ([], [], [])
        self.fake_session.exit_status_ready.side_effect = [0, 0, 0, True]
        self.fake_session.send_ready.return_value = True
        self.fake_session.send.side_effect = [5, 3, 2]
        fake_stdin = mock.Mock()
        fake_stdin.read.side_effect = ["line1", "line2", ""]
        fake_stdin.closed = False

        def close():
            fake_stdin.closed = True
        fake_stdin.close = mock.Mock(side_effect=close)
        self.ssh.run("cmd", stdin=fake_stdin)
        call = mock.call
        send_calls = [call("line1"), call("line2"), call("e2")]
        self.assertEqual(send_calls, self.fake_session.send.mock_calls)

    @mock.patch("rally.utils.sshutils.select")
    def test_run_select_error(self, mock_select):
        self.fake_session.exit_status_ready.return_value = False
        mock_select.select.return_value = ([], [], [True])
        self.assertRaises(exceptions.SSHError, self.ssh.run, "cmd")

    @mock.patch("rally.utils.sshutils.time")
    @mock.patch("rally.utils.sshutils.select")
    def test_run_timemout(self, mock_select, mock_time):
        mock_time.time.side_effect = [1, 3700]
        mock_select.select.return_value = ([], [], [])
        self.fake_session.exit_status_ready.return_value = False
        self.assertRaises(exceptions.SSHTimeout, self.ssh.run, "cmd")

    @mock.patch("rally.utils.sshutils.open", create=True)
    def test__put_file_shell(self, mock_open):
        self.ssh.run = mock.Mock()
        self.ssh._put_file_shell("localfile", "remotefile", 0o42)

        self.ssh.run.assert_called_once_with(
            "cat > remotefile; chmod 042 remotefile",
            stdin=mock_open.return_value.__enter__.return_value)

    @mock.patch("rally.utils.sshutils.os.stat")
    def test__put_file_sftp(self, mock_stat):
        sftp = self.fake_client.open_sftp.return_value = mock.MagicMock()
        sftp.__enter__.return_value = sftp

        mock_stat.return_value = os.stat_result([0o753] + [0] * 9)

        self.ssh._put_file_sftp("localfile", "remotefile")

        sftp.put.assert_called_once_with("localfile", "remotefile")
        mock_stat.assert_called_once_with("localfile")
        sftp.chmod.assert_called_once_with("remotefile", 0o753)
        sftp.__exit__.assert_called_once_with(None, None, None)

    def test__put_file_sftp_mode(self):
        sftp = self.fake_client.open_sftp.return_value = mock.MagicMock()
        sftp.__enter__.return_value = sftp

        self.ssh._put_file_sftp("localfile", "remotefile", mode=0o753)

        sftp.put.assert_called_once_with("localfile", "remotefile")
        sftp.chmod.assert_called_once_with("remotefile", 0o753)
        sftp.__exit__.assert_called_once_with(None, None, None)

    @ddt.data(sshutils.paramiko.SSHException, socket.error)
    def test_put_file(self, exc):
        self.ssh._put_file_sftp = mock.Mock(side_effect=exc())
        self.ssh._put_file_shell = mock.Mock()

        self.ssh.put_file("foo", "bar", 42)
        self.ssh._put_file_sftp.assert_called_once_with("foo", "bar", mode=42)
        self.ssh._put_file_shell.assert_called_once_with("foo", "bar", mode=42)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import alembic
import alembic.config
import alembic.migration
import alembic.script
import sqlalchemy as sa
import sqlalchemy.schema  # noqa

from rally.common.db import api
from rally.common.db import models
from rally import exceptions


INITIAL_REVISION_UUID = "ca3626f62937"


def _alembic_config():
    path = os.path.join(os.path.dirname(__file__), "alembic.ini")
    config = alembic.config.Config(path)
    return config


def schema_cleanup():
    """Drop all database objects.

    Drops all database objects remaining on the default schema of the given
    engine. Per-db implementations will also need to drop items specific to
    those systems, such as sequences, custom types (e.g. pg ENUM), etc.
    """
    engine = api.get_engine()
    with engine.begin() as conn:
        inspector = sa.inspect(engine)
        metadata = sa.schema.MetaData()
        tbs = []
        all_fks = []

        for table_name in inspector.get_table_names():
            fks = []
            for fk in inspector.get_foreign_keys(table_name):
                if not fk["name"]:
                    continue
                fks.append(
                    sa.schema.ForeignKeyConstraint((), (), name=fk["name"]))
            table = sa.schema.Table(table_name, metadata, *fks)
            tbs.append(table)
            all_fks.extend(fks)

        if engine.name != "sqlite":
            for fkc in all_fks:
                conn.execute(sa.schema.DropConstraint(fkc))
        for table in tbs:
            conn.execute(sa.schema.DropTable(table))

        if engine.name == "postgresql":
            sqla_100 = int(sa.__version__.split(".")[0]) >= 1

            if sqla_100:
                enums = [e["name"] for e in sa.inspect(conn).get_enums()]
            else:
                enums = conn.dialect._load_enums(conn).keys()

            for e in enums:
                conn.execute("DROP TYPE %s" % e)


def schema_revision(config=None, engine=None, detailed=False):
    """Current database revision.

    :param config: Instance of alembic config
    :param engine: Instance of DB engine
    :param detailed: whether to return a dict with detailed data
    :rtype detailed: bool
    :returns: Database revision
    :rtype: string
    :rtype: dict
    """
    engine = engine or api.get_engine()
    with engine.connect() as conn:
        context = alembic.migration.MigrationContext.configure(conn)
        revision = context.get_current_revision()
    if detailed:
        config = config or _alembic_config()
        sc_dir = alembic.script.ScriptDirectory.from_config(config)
        return {"revision": revision,
                "current_head": sc_dir.get_current_head()}
    return revision


def schema_upgrade(revision=None, config=None, engine=None):
    """Used for upgrading database.

    :param revision: Desired database version
    :type revision: string
    :param config: Instance of alembic config
    :param engine: Instance of DB engine
    """
    revision = revision or "head"
    config = config or _alembic_config()
    engine = engine or api.get_engine()

    if schema_revision(engine=engine) is None:
        schema_stamp(INITIAL_REVISION_UUID, config=config)

    alembic.command.upgrade(config, revision or "head")


def schema_create(config=None, engine=None):
    """Create database schema from models description.

    Can be used for initial installation instead of upgrade('head').
    :param config: Instance of alembic config
    :param engine: Instance of DB engine
    """
    engine = engine or api.get_engine()

    # NOTE(viktors): If we will use metadata.create_all() for non empty db
    #                schema, it will only add the new tables, but leave
    #                existing as is. So we should avoid of this situation.
    if schema_revision(engine=engine) is not None:
        raise exceptions.DBMigrationError("DB schema is already under version"
                                          " control. Use upgrade() instead")

    models.BASE.metadata.create_all(engine)
    schema_stamp("head", config=config)


def schema_stamp(revision, config=None):
    """Stamps database with provided revision.

    Don't run any migrations.
    :param revision: Should match one from repository or head - to stamp
                     database with most recent revision
    :type revision: string
    :param config: Instance of alembic config
    """
    config = config or _alembic_config()
    return alembic.command.stamp(config, revision)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import subprocess
import sys


DEFAULT_REPORT = ".test_results/%s_report.html" % os.environ["TOX_ENV_NAME"]
PYTEST_REPORT = os.environ.get("PYTEST_REPORT", DEFAULT_REPORT)
PYTEST_ARGUMENTS = ("py.test"  # base command
                    " -vv"  # show test names in logs
                    " --html=%(html_report)s"  # html report
                    " --self-contained-html"  # embedded css
                    " --durations=10"  # get a list of the slowest 10 tests
                    " -n %(concurrency)s"  # launch tests in parallel
                    " %(path)s"
                    )


def error(msg):
    print(msg)
    exit(1)


def main(args):
    parser = argparse.ArgumentParser(args[0])
    parser.add_argument("discovery_path", metavar="<path>", type=str,
                        help="Path to location of all tests.")
    parser.add_argument("--posargs", metavar="<str>", type=str, default="",
                        help="TOX posargs. Currently supported only string to "
                             "partial test or tests group to launch.")
    parser.add_argument("--concurrency", metavar="<N>", type=int,
                        help="Number of parallel processes.")
    args = parser.parse_args(args[1:])

    # We allow only one parameter - path to partial test or tests group
    path = args.posargs
    if len(path.split(" ")) > 1:
        error("Wrong value of posargs. It should include only path to single "
              "test or tests group to launch.")
    # NOTE(andreykurilin): Previously, next format was supported:
    #   tests.unit.test_osclients.SomeTestCase.some_method
    # It is more simple and pythonic than native pytest-way:
    #   tests/unit/test_osclients.py::SomeTestCase::some_method
    # Let's return this support
    if path:
        if "/" not in path:
            path = path.split(".")
            module = ""
            for i in range(0, len(path)):
                part = os.path.join(module, path[i])
                if os.path.exists(part):
                    module = part
                    continue
                if os.path.exists("%s.py" % part):
                    if i != (len(path) - 1):
                        module = "%s.py::%s" % (part, "::".join(path[i + 1:]))
                    else:
                        module = "%s.py" % part
                    break

                error("Non-existing path to single test or tests group to "
                      "launch. %s %s" % (module, part))
            path = module

        path = os.path.abspath(os.path.expanduser(path))
        if not path.startswith(os.path.abspath(args.discovery_path)):
            # Prevent to launch functional tests from unit tests launcher.
            error("Wrong path to single test or tests group to launch. It "
                  "should be in %s." % args.discovery_path)
    else:
        path = args.discovery_path

    print("Test(s) to launch (pytest format): %s" % path)

    args = PYTEST_ARGUMENTS % {"html_report": PYTEST_REPORT,
                               "path": path,
                               "concurrency": args.concurrency or "auto"}
    try:
        subprocess.check_call(args.split(" "),
                              stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError:
        # NOTE(andreykurilin): it is ok, since tests can fail.
        error("")


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for custom sqlalchemy types"""

import mock
import sqlalchemy as sa

from rally.common.db import sa_types
from tests.unit import test


class JsonEncodedTest(test.TestCase):
    def test_impl(self):
        self.assertEqual(sa.Text, sa_types.JSONEncodedDict.impl)
        self.assertEqual(sa.Text, sa_types.JSONEncodedList.impl)
        self.assertEqual(sa.Text, sa_types.MutableJSONEncodedDict.impl)
        self.assertEqual(sa.Text, sa_types.MutableJSONEncodedList.impl)

    def test_process_bind_param(self):
        t = sa_types.JSONEncodedDict()
        self.assertEqual("{\"a\": 1}", t.process_bind_param({"a": 1}, None))

    def test_process_bind_param_none(self):
        t = sa_types.JSONEncodedDict()
        self.assertIsNone(t.process_bind_param(None, None))

    def test_process_result_value(self):
        t = sa_types.JSONEncodedDict()
        self.assertEqual({"a": 1}, t.process_result_value("{\"a\": 1}", None))
        t = sa_types.JSONEncodedList()
        self.assertEqual([[2, 1], [1, 2]], t.process_result_value(
            "[[2, 1], [1, 2]]", None))
        with mock.patch("json.loads") as mock_json_loads:
            t.process_result_value("[[2, 1], [1, 2]]", None)
            mock_json_loads.assert_called_once_with("[[2, 1], [1, 2]]")

    def test_process_result_value_none(self):
        t = sa_types.JSONEncodedDict()
        self.assertIsNone(t.process_result_value(None, None))
        t = sa_types.JSONEncodedList()
        self.assertIsNone(t.process_result_value(None, None))


class MutableDictTest(test.TestCase):
    def test_creation(self):
        sample = {"a": 1, "b": 2}
        d = sa_types.MutableDict(sample)
        self.assertEqual(sample, d)

    def test_coerce_dict(self):
        sample = {"a": 1, "b": 2}
        md = sa_types.MutableDict.coerce("test", sample)
        self.assertEqual(sample, md)
        self.assertIsInstance(md, sa_types.MutableDict)

    def test_coerce_mutable_dict(self):
        sample = {"a": 1, "b": 2}
        sample_md = sa_types.MutableDict(sample)
        md = sa_types.MutableDict.coerce("test", sample_md)
        self.assertEqual(sample, md)
        self.assertIs(sample_md, md)

    def test_coerce_unsupported(self):
        self.assertRaises(ValueError, sa_types.MutableDict.coerce, "test", [])

    @mock.patch.object(sa_types.MutableDict, "changed")
    def test_changed_on_setitem(self, mock_mutable_dict_changed):
        sample = {"a": 1, "b": 2}
        d = sa_types.MutableDict(sample)
        d["b"] = 3
        self.assertEqual({"a": 1, "b": 3}, d)
        self.assertEqual(1, mock_mutable_dict_changed.call_count)

    @mock.patch.object(sa_types.MutableDict, "changed")
    def test_changed_on_delitem(self, mock_mutable_dict_changed):
        sample = {"a": 1, "b": 2}
        d = sa_types.MutableDict(sample)
        del d["b"]
        self.assertEqual({"a": 1}, d)
        self.assertEqual(1, mock_mutable_dict_changed.call_count)


class MutableListTest(test.TestCase):
    def test_creation(self):
        sample = [1, 2, 3]
        d = sa_types.MutableList(sample)
        self.assertEqual(sample, d)

    def test_coerce_list(self):
        sample = [1, 2, 3]
        md = sa_types.MutableList.coerce("test", sample)
        self.assertEqual(sample, md)
        self.assertIsInstance(md, sa_types.MutableList)

    def test_coerce_mutable_list(self):
        sample = [1, 2, 3]
        sample_md = sa_types.MutableList(sample)
        md = sa_types.MutableList.coerce("test", sample_md)
        self.assertEqual(sample, md)
        self.assertIs(sample_md, md)

    def test_coerce_unsupported(self):
        self.assertRaises(ValueError, sa_types.MutableList.coerce, "test", {})

    @mock.patch.object(sa_types.MutableList, "changed")
    def test_changed_on_append(self, mock_mutable_list_changed):
        sample = [1, 2, 3]
        lst = sa_types.MutableList(sample)
        lst.append(4)
        self.assertEqual([1, 2, 3, 4], lst)
        self.assertEqual(1, mock_mutable_list_changed.call_count)

    @mock.patch.object(sa_types.MutableList, "changed")
    def test_changed_on_setitem(self, mock_mutable_list_changed):
        sample = [1, 2, 3]
        lst = sa_types.MutableList(sample)
        lst[2] = 4
        self.assertEqual([1, 2, 4], lst)
        self.assertEqual(1, mock_mutable_list_changed.call_count)

    @mock.patch.object(sa_types.MutableList, "changed")
    def test_changed_on_delitem(self, mock_mutable_list_changed):
        sample = [1, 2, 3]
        lst = sa_types.MutableList(sample)
        del lst[2]
        self.assertEqual([1, 2], lst)
        self.assertEqual(1, mock_mutable_list_changed.call_count)


class TimeStampTestCase(test.TestCase):
    def test_process_bind_param(self):
        self.assertIsNone(sa_types.TimeStamp().process_bind_param(
            None, dialect=None))

        self.assertEqual(
            1498561749348996,
            sa_types.TimeStamp().process_bind_param(1498561749.348996,
                                                    dialect=None))

    def test_process_result_value(self):
        self.assertIsNone(sa_types.TimeStamp().process_result_value(
            None, dialect=None))

        self.assertEqual(
            1498561749.348996,
            sa_types.TimeStamp().process_result_value(1498561749348996,
                                                      dialect=None))



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
SLA (Service-level agreement) is set of details for determining compliance
with contracted values such as maximum error rate or minimum response time.
"""

from rally.common import streaming_algorithms
from rally import consts
from rally.task import sla


@sla.configure(name="outliers")
class Outliers(sla.SLA):
    """Limit the number of outliers (iterations that take too much time).

    The outliers are detected automatically using the computation of the mean
    and standard deviation (std) of the data.
    """
    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA7,
        "properties": {
            "max": {"type": "integer", "minimum": 0},
            "min_iterations": {"type": "integer", "minimum": 3},
            "sigmas": {"type": "number", "minimum": 0.0,
                       "exclusiveMinimum": 0.0}
        },
        "additionalProperties": False,
    }

    def __init__(self, criterion_value):
        super(Outliers, self).__init__(criterion_value)
        self.max_outliers = self.criterion_value.get("max", 0)
        # NOTE(msdubov): Having 3 as default is reasonable (need enough data).
        self.min_iterations = self.criterion_value.get("min_iterations", 3)
        self.sigmas = self.criterion_value.get("sigmas", 3.0)
        self.iterations = 0
        self.outliers = 0
        self.threshold = None
        self.mean_comp = streaming_algorithms.MeanComputation()
        self.std_comp = streaming_algorithms.StdDevComputation()

    def add_iteration(self, iteration):
        # NOTE(ikhudoshyn): This method can not be implemented properly.
        # After adding a new iteration, both mean and standard deviation
        # may change. Hence threshold will change as well. In this case we
        # should again compare durations of all accounted iterations
        # to the threshold. Unfortunately we can not do it since
        # we do not store durations.
        # Implementation provided here only gives rough approximation
        # of outliers number.
        if not iteration.get("error"):
            duration = iteration["duration"]
            self.iterations += 1

            # NOTE(msdubov): First check if the current iteration is an outlier
            if (self.iterations >= self.min_iterations
                    and self.threshold and duration > self.threshold):
                self.outliers += 1

            # NOTE(msdubov): Then update the threshold value
            self.mean_comp.add(duration)
            self.std_comp.add(duration)
            if self.iterations >= 2:
                mean = self.mean_comp.result()
                std = self.std_comp.result()
                self.threshold = mean + self.sigmas * std

        self.success = self.outliers <= self.max_outliers
        return self.success

    def merge(self, other):
        # NOTE(ikhudoshyn): This method can not be implemented properly.
        # After merge, both mean and standard deviation may change.
        # Hence threshold will change as well. In this case we
        # should again compare durations of all accounted iterations
        # to the threshold. Unfortunately we can not do it since
        # we do not store durations.
        # Implementation provided here only gives rough approximation
        # of outliers number.
        self.iterations += other.iterations
        self.outliers += other.outliers
        self.mean_comp.merge(other.mean_comp)
        self.std_comp.merge(other.std_comp)

        if self.iterations >= 2:
            mean = self.mean_comp.result()
            std = self.std_comp.result()
            self.threshold = mean + self.sigmas * std

        self.success = self.outliers <= self.max_outliers
        return self.success

    def details(self):
        return ("Maximum number of outliers %i <= %i - %s" %
                (self.outliers, self.max_outliers, self.status()))



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from subunit import v2

from rally.common import logging
from rally.utils import encodeutils


def prepare_input_args(func):
    # NOTE(andreykurilin): Variables 'runnable', 'eof', 'route_code' are not
    # used in parser.
    def inner(self, test_id=None, test_status=None, timestamp=None,
              file_name=None, file_bytes=None, mime_type=None, test_tags=None,
              runnable=True, eof=False, route_code=None):
        if not test_id:
            return

        if (test_id.startswith("setUpClass (")
                or test_id.startswith("tearDown (")):
            test_id = test_id[test_id.find("(") + 1:-1]

        tags = _parse_test_tags(test_id)

        if mime_type:
            mime_type, charset = mime_type.split("; ")[:2]
            charset = charset.split("=")[1]
        else:
            charset = None

        func(self, test_id, test_status, timestamp, tags,
             file_name, file_bytes, test_tags, mime_type, charset)

    return inner


def _parse_test_tags(test_id):
    tags = []
    if test_id.find("[") > -1:
        tags = test_id.split("[")[1][:-1].split(",")

    return tags


class SubunitV2StreamResult(object):

    def __init__(self, expected_failures=None, skipped_tests=None, live=False,
                 logger_name=None):
        self._tests = {}
        self._expected_failures = expected_failures or {}
        self._skipped_tests = skipped_tests or {}

        self._live = live
        self._logger = logging.getLogger(logger_name or __name__)

        self._timestamps = {}
        # NOTE(andreykurilin): _first_timestamp and _last_timestamp variables
        # are designed to calculate the total time of tests execution.
        self._first_timestamp = None
        self._last_timestamp = None

        # Store unknown entities and process them later.
        self._unknown_entities = {}
        self._is_parsed = False

    @staticmethod
    def _get_test_name(test_id):
        return test_id.split("[")[0] if test_id.find("[") > -1 else test_id

    def _check_expected_failure(self, test_id):
        if (test_id in self._expected_failures
                or self._get_test_name(test_id) in self._expected_failures):
            if self._tests[test_id]["status"] == "fail":
                self._tests[test_id]["status"] = "xfail"
                if self._expected_failures[test_id]:
                    self._tests[test_id]["reason"] = (
                        self._expected_failures[test_id])
            elif self._tests[test_id]["status"] == "success":
                self._tests[test_id]["status"] = "uxsuccess"

    def _process_skipped_tests(self):
        for t_id in self._skipped_tests.copy():
            if t_id not in self._tests:
                status = "skip"
                name = self._get_test_name(t_id)
                self._tests[t_id] = {"status": status,
                                     "name": name,
                                     "duration": "%.3f" % 0,
                                     "tags": _parse_test_tags(t_id)}
                if self._skipped_tests[t_id]:
                    self._tests[t_id]["reason"] = self._skipped_tests[t_id]
                    status += ": %s" % self._tests[t_id]["reason"]
                if self._live:
                    self._logger.info("{-} %s ... %s" % (name, status))

            self._skipped_tests.pop(t_id)

    def _parse(self):
        # NOTE(andreykurilin): When whole test class is marked as skipped or
        # failed, there is only one event with reason and status. So we should
        # modify all tests of test class manually.
        for test_id in self._unknown_entities:
            known_test_ids = filter(
                lambda t: t == test_id or t.startswith("%s." % test_id),
                self._tests)
            for t_id in known_test_ids:
                if self._tests[t_id]["status"] == "init":
                    self._tests[t_id]["status"] = (
                        self._unknown_entities[test_id]["status"])

                if self._unknown_entities[test_id].get("reason"):
                    self._tests[t_id]["reason"] = (
                        self._unknown_entities[test_id]["reason"])
                elif self._unknown_entities[test_id].get("traceback"):
                    self._tests[t_id]["traceback"] = (
                        self._unknown_entities[test_id]["traceback"])

        # decode data
        for test_id in self._tests:
            for file_name in ["traceback", "reason"]:
                # TODO(andreykurilin): decode fields based on mime_type
                if file_name in self._tests[test_id]:
                    self._tests[test_id][file_name] = (
                        encodeutils.safe_decode(
                            self._tests[test_id][file_name]))

        self._is_parsed = True

    @property
    def tests(self):
        if not self._is_parsed:
            self._parse()
        return self._tests

    @property
    def totals(self):
        td = 0
        if self._first_timestamp:
            td = (self._last_timestamp - self._first_timestamp).total_seconds()

        return {"tests_count": len(self.tests),
                "tests_duration": "%.3f" % td,
                "failures": len(self.filter_tests("fail")),
                "skipped": len(self.filter_tests("skip")),
                "success": len(self.filter_tests("success")),
                "unexpected_success": len(self.filter_tests("uxsuccess")),
                "expected_failures": len(self.filter_tests("xfail"))}

    @prepare_input_args
    def status(self, test_id=None, test_status=None, timestamp=None, tags=None,
               file_name=None, file_bytes=None, worker=None, mime_type=None,
               charset=None):
        if timestamp:
            if not self._first_timestamp:
                self._first_timestamp = timestamp
            self._last_timestamp = timestamp

        if test_status == "exists":
            self._tests[test_id] = {"status": "init",
                                    "name": self._get_test_name(test_id),
                                    "duration": "%.3f" % 0,
                                    "tags": tags if tags else []}
        elif test_id in self._tests:
            if test_status == "inprogress":
                # timestamp of test start
                self._timestamps[test_id] = timestamp
                self._tests[test_id]["timestamp"] = timestamp.strftime(
                    "%Y-%m-%dT%H:%M:%S%z")
            elif test_status:
                self._tests[test_id]["duration"] = "%.3f" % (
                    timestamp - self._timestamps[test_id]).total_seconds()
                self._tests[test_id]["status"] = test_status

                self._check_expected_failure(test_id)
            else:
                if file_name in ["traceback", "reason"]:
                    if file_name not in self._tests[test_id]:
                        self._tests[test_id][file_name] = file_bytes
                    else:
                        self._tests[test_id][file_name] += file_bytes
        else:
            self._unknown_entities.setdefault(test_id, {"name": test_id})
            self._unknown_entities[test_id]["status"] = test_status
            if file_name in ["traceback", "reason"]:
                if file_name not in self._unknown_entities[test_id]:
                    self._unknown_entities[test_id][file_name] = file_bytes
                else:
                    self._unknown_entities[test_id][file_name] += file_bytes

        if self._skipped_tests:
            self._process_skipped_tests()

        if self._live and test_status not in (None, "exists", "inprogress"):
            duration = ""
            if test_id in self._tests:
                status = self._tests[test_id]["status"]
                duration = " [%ss]" % self._tests[test_id]["duration"]
            else:
                status = test_status

            status += duration

            if "xfail" in status or "skip" in status:
                if test_id in self._tests:
                    reason = self._tests[test_id].get("reason")
                else:
                    reason = self._unknown_entities[test_id].get("reason")
                if reason:
                    status += ": %s" % reason

            w = "{%s} " % worker.pop().split("-")[1] if worker else "-"
            self._logger.info("%s ... %s"
                              % (w + self._get_test_name(test_id), status))

    def filter_tests(self, status):
        """Filter tests by given status."""
        filtered_tests = {}
        for test in self.tests:
            if self.tests[test]["status"] == status:
                filtered_tests[test] = self.tests[test]

        return filtered_tests


def parse(stream, expected_failures=None, skipped_tests=None, live=False,
          logger_name=None):
    results = SubunitV2StreamResult(expected_failures, skipped_tests, live,
                                    logger_name)
    v2.ByteStreamToStreamResult(stream, "non-subunit").run(results)

    return results


def parse_file(filename, expected_failures=None, skipped_tests=None,
               live=False, logger_name=None):
    with open(filename, "rb") as stream:
        return parse(stream, expected_failures, skipped_tests, live,
                     logger_name)



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from rally.plugins.common.sla import max_average_duration_per_atomic as madpa
from rally.task import sla
from tests.unit import test


@ddt.ddt
class MaxAverageDurationPerAtomicTestCase(test.TestCase):

    @ddt.data(({"a": 10, "b": 20}, True),
              ({"a": "foo"}, False),
              ({}, False))
    @ddt.unpack
    def test_validate(self, config, valid):
        results = sla.SLA.validate(
            "max_avg_duration_per_atomic", None, None, config)
        if valid:
            self.assertEqual([], results)
        else:
            self.assertEqual(1, len(results))

    def test_result(self):
        cls = madpa.MaxAverageDurationPerAtomic
        sla1 = cls({"a1": 42, "a2": 42})
        sla2 = cls({"a1": 42, "a2": 2})
        for sla_inst in [sla1, sla2]:
            sla_inst.add_iteration(
                {"atomic_actions": [
                    {"name": "a1",
                     "started_at": 0.0,
                     "finished_at": 3.14},
                    {"name": "a2",
                     "started_at": 10,
                     "finished_at": 17.77}]})
            sla_inst.add_iteration(
                {"atomic_actions": [
                    {"name": "a1",
                     "started_at": 12.0,
                     "finished_at": 15.14},
                    {"name": "a2",
                     "started_at": 13,
                     "finished_at": 27.33}]})
        self.assertTrue(sla1.result()["success"])
        self.assertFalse(sla2.result()["success"])
        self.assertEqual("Passed", sla1.status())
        self.assertEqual("Failed", sla2.status())

    def test_result_no_iterations(self):
        sla_inst = madpa.MaxAverageDurationPerAtomic({"a1": 8.14, "a2": 9.77})
        self.assertTrue(sla_inst.result()["success"])

    def test_add_iteration(self):
        sla_inst = madpa.MaxAverageDurationPerAtomic({"a1": 5, "a2": 10})
        add = sla_inst.add_iteration
        self.assertTrue(add({"atomic_actions": [
            {"name": "a1", "started_at": 0, "finished_at": 2.5},
            {"name": "a2", "started_at": 0, "finished_at": 5.0}]}))
        self.assertTrue(add({"atomic_actions": [
            {"name": "a1", "started_at": 0, "finished_at": 5.0},
            {"name": "a2", "started_at": 0, "finished_at": 10.0}]}))
        # the following pushes a2 over the limit
        self.assertFalse(add({"atomic_actions": [
            {"name": "a1", "started_at": 0, "finished_at": 5.0},
            {"name": "a2", "started_at": 0, "finished_at": 20.0}]}))
        # bring a2 back
        self.assertTrue(add({"atomic_actions": [
            {"name": "a1", "started_at": 0, "finished_at": 5.0},
            {"name": "a2", "started_at": 0, "finished_at": 2.0}]}))
        # push a1 over
        self.assertFalse(add({"atomic_actions": [
            {"name": "a1", "started_at": 0, "finished_at": 10.0},
            {"name": "a2", "started_at": 0, "finished_at": 2.0}]}))
        # bring it back
        self.assertTrue(add({"atomic_actions": [
            {"name": "a1", "started_at": 0, "finished_at": 1.0},
            {"name": "a2", "started_at": 0, "finished_at": 2.0}]}))

    def test_merge(self):
        durations = [[1.0, 2.0, 1.5, 4.3],
                     [2.1, 3.4, 1.2, 6.3, 7.2, 7.0, 1.],
                     [1.1, 1.1, 2.2, 2.2, 3.3, 4.3]]
        init = {"a1": 8.14, "a2": 9.77}
        single_sla = madpa.MaxAverageDurationPerAtomic(init)

        for dd in durations:
            for d in dd:
                single_sla.add_iteration(
                    {"atomic_actions": [{"name": "a1",
                                         "started_at": 0,
                                         "finished_at": d},
                                        {"name": "a2",
                                         "started_at": d,
                                         "finished_at": d * 3}
                                        ]})

        slas = [madpa.MaxAverageDurationPerAtomic(init) for _ in durations]

        for idx, sla_inst in enumerate(slas):
            for d in durations[idx]:
                sla_inst.add_iteration(
                    {"atomic_actions": [{"name": "a1",
                                         "started_at": 0,
                                         "finished_at": d},
                                        {"name": "a2",
                                         "started_at": d,
                                         "finished_at": d * 3}
                                        ]})

        merged_sla = slas[0]
        for sla_inst in slas[1:]:
            merged_sla.merge(sla_inst)

        self.assertEqual(single_sla.success, merged_sla.success)
        self.assertEqual(single_sla.avg_by_action, merged_sla.avg_by_action)



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import mock
from six import moves

from rally.cli import envutils
from rally import exceptions
from tests.unit import test


class EnvUtilsTestCase(test.TestCase):

    def test_default_from_global(self):

        @envutils.default_from_global("test_arg_name",
                                      "test_env_name",
                                      "test_missing_arg")
        def test_function(test_arg_name=None):
            pass

        with mock.patch("sys.stdout",
                        new_callable=moves.StringIO) as mock_stdout:
            test_function()
            self.assertEqual("Missing argument: --test_missing_arg\n",
                             mock_stdout.getvalue())

    @mock.patch.dict(os.environ,
                     values={envutils.ENV_DEPLOYMENT: "my_deployment_id"},
                     clear=True)
    def test_get_deployment_id_in_env(self):
        deployment_id = envutils.get_global(envutils.ENV_DEPLOYMENT)
        self.assertEqual("my_deployment_id", deployment_id)

    @mock.patch.dict(os.environ, values={}, clear=True)
    @mock.patch("rally.cli.envutils.fileutils.load_env_file")
    def test_get_deployment_id_with_exception(self, mock_load_env_file):
        self.assertRaises(exceptions.InvalidArgumentsException,
                          envutils.get_global, envutils.ENV_DEPLOYMENT, True)
        mock_load_env_file.assert_called_once_with(os.path.expanduser(
            "~/.rally/globals"))

    @mock.patch.dict(os.environ, values={}, clear=True)
    @mock.patch("rally.cli.envutils.fileutils.load_env_file")
    def test_get_deployment_id_with_none(self, mock_load_env_file):
        self.assertIsNone(envutils.get_global(envutils.ENV_DEPLOYMENT))
        mock_load_env_file.assert_called_once_with(os.path.expanduser(
            "~/.rally/globals"))

    @mock.patch.dict(os.environ, values={envutils.ENV_TASK: "my_task_id"},
                     clear=True)
    def test_get_task_id_in_env(self):
        self.assertEqual("my_task_id", envutils.get_global(envutils.ENV_TASK))

    @mock.patch.dict(os.environ, values={}, clear=True)
    @mock.patch("rally.cli.envutils.fileutils.load_env_file")
    def test_get_task_id_with_exception(self, mock_load_env_file):
        self.assertRaises(exceptions.InvalidArgumentsException,
                          envutils.get_global, envutils.ENV_TASK, True)
        mock_load_env_file.assert_called_once_with(os.path.expanduser(
            "~/.rally/globals"))

    @mock.patch.dict(os.environ, values={}, clear=True)
    @mock.patch("rally.cli.envutils.fileutils.load_env_file")
    def test_get_task_id_with_none(self, mock_load_env_file):
        self.assertIsNone(envutils.get_global("RALLY_TASK"))
        mock_load_env_file.assert_called_once_with(os.path.expanduser(
            "~/.rally/globals"))

    @mock.patch.dict(os.environ,
                     values={envutils.ENV_DEPLOYMENT: "test_deployment_id"},
                     clear=True)
    @mock.patch("os.path.exists")
    @mock.patch("rally.cli.envutils.fileutils.update_env_file",
                return_value=True)
    def test_clear_global(self, mock_update_env_file, mock_path_exists):
        envutils.clear_global(envutils.ENV_DEPLOYMENT)
        mock_update_env_file.assert_called_once_with(os.path.expanduser(
            "~/.rally/globals"), envutils.ENV_DEPLOYMENT, "\n")
        self.assertEqual({}, os.environ)

    @mock.patch.dict(os.environ,
                     values={envutils.ENV_DEPLOYMENT: "test_deployment_id",
                             envutils.ENV_TASK: "test_task_id"},
                     clear=True)
    @mock.patch("os.path.exists")
    @mock.patch("rally.cli.envutils.fileutils.update_env_file",
                return_value=True)
    def test_clear_env(self, mock_update_env_file, mock_path_exists):
        envutils.clear_env()
        self.assertEqual({}, os.environ)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import itertools
import os
import re

from tests.unit import test


class Variants(object):
    def __init__(self, variants, print_prefix="mock_"):
        self.variants = variants
        self.print_prefix = print_prefix

    def __repr__(self):
        variants = self.variants
        if len(variants) > 3:
            variants = variants[:3]
        variants = [repr(self.print_prefix + var) for var in variants]
        return "{" + ", ".join(variants) + (
            ", ...}" if len(self.variants) > 3 else "}")

    def __eq__(self, val):
        return getattr(val, "variants", val) == self.variants

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, val):
        return val in self.variants


def pairwise_isinstance(*args):
    return all(itertools.starmap(isinstance, args))


class FuncMockArgsDecoratorsChecker(ast.NodeVisitor):
    """Recursively visit an AST looking for misusage of mocks in tests.

    The misusage being tested by this particular class is unmatched mocked
    object name against the argument names.

    The following is the correct usages::
        @mock.patch("module.abc") # or
        # or @mock.patch(MODULE + ".abc")
        # or @mock.patch("%s.abc" % MODULE) where MODULE="module"
        def test_foobar(self, mock_module_abc): # or `mock_abc'
            ...

        @mock.patch("pkg.ClassName.abc") # or
        # or @mock.patch(CLASSNAME + ".abc")
        # or @mock.patch("%s.abc" % CLASSNAME) where CLASSNAME="pkg.ClassName"
        def test_foobar(self, mock_class_name_abc):
            ...

        class FooClassNameTestCase(...):
            @mock.patch("pkg.FooClassName.abc")
            def test_foobar(self, mock_abc):
                # Iff the mocked object is inside the tested class then
                # the class name in mock argname is optional.
                ...

    While these are not::
        @mock.patch("module.abc")
        def test_foobar(self, m_abc):
            # must be prefixed with `mock_'

        @mock.patch("module.abc")
        def test_foobar(self, mock_cba):
            # must contain mocked object name (`mock_abc')

        @mock.patch("module.abc")
        def test_foobar(self, mock_modulewrong_abc):
            # must match the module `mock_module_abc'

        @mock.patch("ClassName.abc")
        def test_foobar(self, mock_class_abc):
            # must match the python-styled class name + method name
    """

    # NOTE(amaretskiy): Disable check if shortest variant is too long
    #                   because long name is not convenient and could
    #                   even be blocked by PEP8
    SHORTEST_VARIANT_LEN_LIMIT = 25

    def __init__(self):
        self.errors = []
        self.globals_ = {}

    @classmethod
    def _get_name(cls, node):
        if isinstance(node, ast.Name):
            return node.id
        if isinstance(node, ast.Attribute):
            return cls._get_name(node.value) + "." + node.attr
        return ""

    def _get_value(self, node):
        """Get mock.patch string argument regexp.

        It is either a string (if we are lucky), string-format of
        ("%s.something" % GVAL) or (GVAL + ".something")
        """
        val = None
        if isinstance(node, ast.Str):
            val = node.s
        elif isinstance(node, ast.BinOp):
            if pairwise_isinstance(
                    (node.op, ast.Mod), (node.left, ast.Str),
                    (node.right, ast.Name)):
                val = node.left.s % self.globals_[node.right.id]
            elif pairwise_isinstance(
                    (node.op, ast.Add), (node.left, ast.Name),
                    (node.right, ast.Str)):
                val = self.globals_[node.left.id] + node.right.s
        elif isinstance(node, ast.Name):
            val = self.globals_[node.id]

        if val is None:
            raise ValueError(
                "Unable to find value in %s, only the following are parsed: "
                "GLOBAL, 'pkg.foobar', '%%s.foobar' %% GLOBAL or 'GLOBAL + "
                "'.foobar'"
                % ast.dump(node))

        return val

    CAMELCASE_SPLIT_ANY_AND_CAPITAL = re.compile("(.)([A-Z][a-z]+)")
    CAMELCASE_SPLIT_LOWER_AND_CAPITAL = re.compile("([a-z0-9])([A-Z])")
    CAMELCASE_SPLIT_REPL = r"\1_\2"

    @classmethod
    def _camelcase_to_python(cls, name):
        for regexp in (cls.CAMELCASE_SPLIT_ANY_AND_CAPITAL,
                       cls.CAMELCASE_SPLIT_LOWER_AND_CAPITAL):
            name = regexp.sub(cls.CAMELCASE_SPLIT_REPL, name)
        return name.lower()

    def _get_mocked_class_value_variants(self, class_name, mocked_name):
        class_name = self._camelcase_to_python(class_name)
        mocked_name = self._camelcase_to_python(mocked_name)

        if class_name == self.classname_python:
            # Optional, since class name of the mocked package is the same as
            # class name of the *TestCase
            return [mocked_name, class_name + "_" + mocked_name]

        # Full class name is required otherwise
        return [class_name + "_" + mocked_name]

    def _add_pkg_optional_prefixes(self, tokens, variants):
        prefixed_variants = list(variants)
        for token in map(self._camelcase_to_python, reversed(tokens)):
            prefixed_variants.append(token + "_" + prefixed_variants[-1])
        return prefixed_variants

    def _get_mocked_name_variants(self, name):
        tokens = name.split(".")
        variants = [self._camelcase_to_python(tokens.pop())]
        if tokens:
            if tokens[-1][0].isupper():
                # Mocked something inside a class, check if we should require
                # the class name to be present in mock argument
                variants = self._get_mocked_class_value_variants(
                    class_name=tokens.pop(),
                    mocked_name=variants[0])
            variants = self._add_pkg_optional_prefixes(tokens, variants)
        return Variants(variants)

    def _get_mock_decorators_variants(self, funccall):
        """Return all the mock.patch{,.object} decorated for function."""
        mock_decorators = []

        for decorator in reversed(funccall.decorator_list):
            if not isinstance(decorator, ast.Call):
                continue
            funcname = self._get_name(decorator.func)

            if funcname == "mock.patch":
                decname = self._get_value(decorator.args[0])
            elif funcname == "mock.patch.object":
                decname = (self._get_name(decorator.args[0]) + "."
                           + self._get_value(decorator.args[1]))
            else:
                continue

            mock_decorators.append(
                self._get_mocked_name_variants(decname)
            )

        return mock_decorators

    @staticmethod
    def _get_mock_args(node):
        """Return all the mock arguments."""
        args = []
        PREFIX_LENGTH = len("mock_")

        for arg in node.args.args:
            name = getattr(arg, "id", getattr(arg, "arg", None))
            if not name.startswith("mock_"):
                continue
            args.append(name[PREFIX_LENGTH:])

        return args

    def visit_Assign(self, node):
        """Catch all the globals."""
        self.generic_visit(node)

        if node.col_offset == 0:
            mnode = ast.Module(body=[node])
            code = compile(mnode, "<ast>", "exec")
            try:
                exec(code, self.globals_)
            except Exception:
                pass
        self.globals_.pop("__builtins__", None)
        self.globals_.pop("builtins", None)

    def visit_ClassDef(self, node):
        classname_camel = node.name
        if node.name.endswith("TestCase"):
            classname_camel = node.name[:-len("TestCase")]

        self.classname_python = self._camelcase_to_python(classname_camel)

        self.generic_visit(node)

    def check_name(self, arg, dec_vars):
        return (dec_vars is not None and arg in dec_vars)

    def visit_FunctionDef(self, node):
        self.generic_visit(node)

        mock_decs = self._get_mock_decorators_variants(node)

        if not mock_decs:
            return

        mock_args = self._get_mock_args(node)

        error_msgs = []
        mismatched = False
        for arg, dec_vars in itertools.zip_longest(mock_args, mock_decs):
            if not self.check_name(arg, dec_vars):
                if arg and dec_vars:
                    sorted_by_len = sorted(
                        dec_vars.variants, key=lambda i: len(i), reverse=True)
                    shortest_name = sorted_by_len.pop()
                    if len(shortest_name) <= self.SHORTEST_VARIANT_LEN_LIMIT:
                        error_msgs.append(
                            ("Argument '%(arg)s' misnamed; should be either "
                             "of %(dec)s that is derived from the mock "
                             "decorator args.\n") % {"arg": arg,
                                                     "dec": dec_vars})
                elif not arg:
                    error_msgs.append(
                        "Missing or malformed argument for %s decorator."
                        % dec_vars)
                    mismatched = True
                elif not dec_vars:
                    error_msgs.append(
                        "Missing or malformed decorator for '%s' argument."
                        % arg)
                    mismatched = True

        if error_msgs:
            if mismatched:
                self.errors.append({
                    "lineno": node.lineno,
                    "args": mock_args,
                    "decs": mock_decs,
                    "messages": error_msgs
                })
            else:
                self.errors.append({
                    "lineno": node.lineno,
                    "mismatch_pairs": list(zip(mock_args, mock_decs)),
                    "messages": error_msgs
                })


class MockUsageCheckerTestCase(test.TestCase):
    tests_path = os.path.join(os.path.dirname(__file__))

    def test_mock_decorators_and_args(self):
        """Ensure that mocked objects are called correctly in the arguments.

        See `FuncMockArgsDecoratorsChecker' docstring for details.
        """
        errors = []

        for dirname, dirnames, filenames in os.walk(self.tests_path):
            for filename in filenames:
                if (not filename.startswith("test_")
                        or not filename.endswith(".py")):
                    continue

                filename = os.path.relpath(os.path.join(dirname, filename))

                with open(filename, "rb") as fh:
                    tree = ast.parse(fh.read(), filename)

                visitor = FuncMockArgsDecoratorsChecker()
                visitor.visit(tree)
                errors.extend(
                    dict(filename=filename, **error)
                    for error in visitor.errors)

        if errors:
            print(FuncMockArgsDecoratorsChecker.__doc__)
            print(
                "\n\n"
                "The following errors were found during the described check:")
            for error in errors:
                print("\n\n"
                      "Errors at file %(filename)s line %(lineno)d:\n\n"
                      "%(message)s" % {
                          "message": "\n".join(error["messages"]),
                          "filename": error["filename"],
                          "lineno": error["lineno"]})

        # NOTE(pboldin): When the STDOUT is shuted the below is the last
        # resort to know what is wrong with the mock names.
        for error in errors:
            error["messages"] = [
                message.rstrip().replace("\n", "  ").replace("\t", "")
                for message in error["messages"]
            ]
        self.assertEqual([], errors)



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from docutils import nodes

from rally.common.plugin import plugin
from rally.common import validation
from rally import plugins
from rally.task import scenario
from tests.unit.doc import utils
from tests.unit import test


class DocstringsTestCase(test.TestCase):

    def setUp(self):
        super(DocstringsTestCase, self).setUp()
        plugins.load()

    def _validate_code_block(self, plg_cls, code_block):
        ignored_params = ["self", "scenario_obj"]
        params_count = code_block.co_argcount
        params = code_block.co_varnames[:params_count]
        param_data = plg_cls.get_info()["parameters"]
        documented_params = [p["name"] for p in param_data]
        result = []
        for param in params:
            if param not in ignored_params:
                if param not in documented_params:
                    msg = ("Class: %(class)s Docstring for "
                           "%(scenario)s should"
                           " describe the '%(param)s' parameter"
                           " in the :param <name>: clause."
                           % {"class": plg_cls.__name__,
                              "scenario": plg_cls.get_name(),
                              "param": param})
                    result.append(msg)
        return result

    # the list with plugins names which use rst definitions in their docstrings
    _HAS_VALID_DEFINITIONS = []

    def _iterate_parsed_rst(self, plugin_name, items, msg_buffer):
        for item in items:
            if (isinstance(item, nodes.definition_list)
                    and plugin_name not in self._HAS_VALID_DEFINITIONS):
                msg_buffer.append("Plugin %s has a docstring with invalid "
                                  "format. Re-check intend and required empty "
                                  "lines between the list title and list "
                                  "items." % plugin_name)
            elif isinstance(item, nodes.system_message):
                msg_buffer.append(
                    "A warning is caught while parsing docstring of '%s' "
                    "plugin: %s" % (plugin_name, item.astext()))
            elif item.children:
                self._iterate_parsed_rst(plugin_name, item.children,
                                         msg_buffer)

    def _check_docstrings(self, msg_buffer):
        for plg_cls in plugin.Plugin.get_all():
            if not plg_cls.__module__.startswith("rally."):
                continue
            name = "%s (%s.%s)" % (plg_cls.get_name(),
                                   plg_cls.__module__,
                                   plg_cls.__name__)
            doc_info = plg_cls.get_info()
            if not doc_info["title"]:
                msg_buffer.append("Plugin '%s' should have a docstring."
                                  % name)
            if doc_info["title"].startswith("Test"):
                msg_buffer.append("One-line description for %s"
                                  " should be declarative and not"
                                  " start with 'Test(s) ...'"
                                  % name)

            # NOTE(andreykurilin): I never saw any real usage of
            #   reStructuredText definitions in our docstrings. In most cases,
            #   "definitions" means that there is an issue with intends or
            #   missed empty line before the list title and list items.
            if doc_info["description"]:
                parsed_docstring = utils.parse_rst(doc_info["description"])
                self._iterate_parsed_rst(plg_cls.get_name(),
                                         parsed_docstring,
                                         msg_buffer)

    def _check_described_params(self, msg_buffer):
        for plg_cls in plugin.Plugin.get_all():
            msg = []
            if hasattr(plg_cls, "run") and issubclass(
                    plg_cls, scenario.Scenario):
                msg = self._validate_code_block(plg_cls,
                                                plg_cls.run.__code__)
            elif hasattr(plg_cls, "validate") and issubclass(
                    plg_cls, validation.Validator):
                msg = self._validate_code_block(plg_cls,
                                                plg_cls.__init__.__code__)
            msg_buffer.extend(msg) if len(msg) else None

    def test_all_plugins_have_docstrings(self):
        msg_buffer = []
        self._check_docstrings(msg_buffer)

        self._check_described_params(msg_buffer)
        if msg_buffer:
            self.fail("\n%s" % "\n===============\n".join(msg_buffer))

    def test_plugin_bases_have_docstrigs(self):
        plugin_bases = set()
        msg_buffer = []
        for plg_cls in plugin.Plugin.get_all(allow_hidden=True):
            plugin_bases.add(plg_cls._get_base())
        for base in plugin_bases:
            name = "%s.%s" % (base.__module__, base.__name__)
            try:
                docstring = base._get_doc()
            except Exception:
                docstring = base.__doc__

            print(name)
            parsed_docstring = utils.parse_rst(docstring)
            self._iterate_parsed_rst(name,
                                     parsed_docstring,
                                     msg_buffer)

        if msg_buffer:
            self.fail("\n%s" % "\n===============\n".join(msg_buffer))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import mock

from rally import exceptions
from rally.task import task_cfg
from tests.unit import test


class TaskConfigTestCase(test.TestCase):

    def test_init_empty_config(self):
        config = None
        e = self.assertRaises(exceptions.InvalidTaskException,
                              task_cfg.TaskConfig, config)
        self.assertEqual("It is empty", e.kwargs["message"])

    def test_validate_version(self):
        p1001 = mock.Mock()
        p1002 = mock.Mock()

        valid_cfg = {"version": 0, "title": "", "subtasks": []}

        class TaskConfig(task_cfg.TaskConfig):
            def _process_1001(self, config):
                p1001()
                return valid_cfg

            def _process_1002(self, config):
                p1002()
                return valid_cfg

        TaskConfig({"version": 1001})
        p1001.assert_called_once_with()
        self.assertFalse(p1002.called)

        p1001.reset_mock()

        TaskConfig({"version": 1002})
        p1002.assert_called_once_with()
        self.assertFalse(p1001.called)

    def test__process_1(self):

        config = collections.OrderedDict()
        config["a.task"] = [{"context": {"foo": "bar"}}, {}]
        config["b.task"] = [{"sla": {"key": "value"}}]
        config["c.task"] = [{"hooks": [{"name": "foo",
                                        "args": "bar",
                                        "description": "DESCR!!!",
                                        "trigger": {
                                            "name": "mega-trigger",
                                            "args": {"some": "thing"}
                                        }}]
                             }]
        self.assertEqual(
            {"title": "Task (adopted from task format v1)",
             "version": 2,
             "description": "",
             "tags": [],
             "subtasks": [
                 {
                     "title": "a.task",
                     "description": "",
                     "tags": [],
                     "workloads": [
                         {
                             "scenario": {"a.task": {}},
                             "contexts": {"foo": "bar"},
                             "hooks": [],
                             "sla": {"failure_rate": {"max": 0}},
                             "runner": {"serial": {}}
                         },
                         {
                             "scenario": {"a.task": {}},
                             "contexts": {},
                             "hooks": [],
                             "sla": {"failure_rate": {"max": 0}},
                             "runner": {"serial": {}}
                         }
                     ],
                 },
                 {
                     "title": "b.task",
                     "description": "",
                     "tags": [],
                     "workloads": [
                         {
                             "scenario": {"b.task": {}},
                             "contexts": {},
                             "hooks": [],
                             "runner": {"serial": {}},
                             "sla": {"key": "value"},
                         }
                     ],
                 },
                 {
                     "title": "c.task",
                     "description": "",
                     "tags": [],
                     "workloads": [
                         {
                             "scenario": {"c.task": {}},
                             "contexts": {},
                             "hooks": [{
                                 "description": "DESCR!!!",
                                 "action": {"foo": "bar"},
                                 "trigger": {
                                     "mega-trigger": {"some": "thing"}}}
                             ],
                             "sla": {"failure_rate": {"max": 0}},
                             "runner": {"serial": {}}
                         }
                     ],
                 }]},
            task_cfg.TaskConfig(config).to_dict())

    def test__process_2(self):

        config = {
            "version": 2,
            "title": "foo",
            "subtasks": [
                {
                    "title": "subtask1",
                    "workloads": [
                        {
                            "scenario": {"workload1": {}},
                            "runner": {"constant": {}},
                            "sla": {"key": "value"}
                        },
                        {
                            "scenario": {"workload2": {}},
                        }
                    ]
                },
                {
                    "title": "subtask2",
                    "scenario": {"workload1": {}}
                },
            ]
        }

        self.assertEqual(
            {"title": "foo",
             "version": 2,
             "description": "",
             "tags": [],
             "subtasks": [
                 {
                     "title": "subtask1",
                     "description": "",
                     "tags": [],
                     "workloads": [
                         {
                             "scenario": {"workload1": {}},
                             "contexts": {},
                             "hooks": [],
                             "sla": {"key": "value"},
                             "runner": {"constant": {}}
                         },
                         {
                             "scenario": {"workload2": {}},
                             "contexts": {},
                             "hooks": [],
                             "sla": {"failure_rate": {"max": 0}},
                             "runner": {"serial": {}}
                         }
                     ],
                 },
                 {
                     "title": "subtask2",
                     "description": "",
                     "tags": [],
                     "workloads": [
                         {
                             "scenario": {"workload1": {}},
                             "contexts": {},
                             "hooks": [],
                             "runner": {"serial": {}},
                             "sla": {"failure_rate": {"max": 0}},
                         }
                     ],
                 }]},
            task_cfg.TaskConfig(config).to_dict())

    def test_hook_config_compatibility(self):
        cfg = {
            "xxx": [{
                "args": {},
                "runner": {"type": "yyy"},
                "hooks": [
                    {
                        "description": "descr",
                        "name": "hook_action",
                        "args": {"k1": "v1"},
                        "trigger": {
                            "name": "hook_trigger",
                            "args": {"k2": "v2"}
                        }
                    }
                ]
            }]
        }
        task = task_cfg.TaskConfig(cfg)
        workload = task.subtasks[0]["workloads"][0]
        self.assertEqual(
            {"description": "descr",
             "action": ("hook_action", {"k1": "v1"}),
             "trigger": ("hook_trigger", {"k2": "v2"})},
            workload["hooks"][0])

    # NOTE(andreykurilin): For a long time, we had a single JSON Schema for
    #   checking the task config. It was so complex that validation errors were
    #   unable to say anything helpful to end-users. The purpose of the
    #   following negative tests is to ensure that UX is on the good level.

    def test_validate_wrong_version(self):
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 5})
        self.assertEqual(
            "Task configuration version 5 is not supported. "
            "Supported versions: 1, 2",
            e.kwargs["message"])

    def test_v2_invalid_top_level(self):
        # single missed property
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "subtasks": []})
        self.assertEqual(
            "'title' is a required property, but it is missed.",
            e.kwargs["message"])
        # multiple missed properties (msg should be different)
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2})
        self.assertEqual(
            "'subtasks', 'title' are required properties, but they are "
            "missed.", e.kwargs["message"])

        # single redundant property
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "title": "", "subtasks": [],
                                  "foo": "bar"})
        self.assertEqual(
            "Additional properties are not allowed ('foo' was unexpected).",
            e.kwargs["message"]
        )
        # multiple redundant properties (msg should be different)
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "title": "", "subtasks": [],
                                  "foo": "bar", "xxx": "yyy"})
        self.assertEqual(
            "Additional properties are not allowed ('foo', 'xxx' were "
            "unexpected).",
            e.kwargs["message"]
        )

    def test_v2_title(self):
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "title": {}, "subtasks": []})
        self.assertEqual(
            "Title should be a string, but 'dict' is found.",
            e.kwargs["message"]
        )

        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "title": "a" * 300,
                                  "subtasks": []})
        self.assertEqual(
            "Title should not be longer then 254 char. Use 'description' field"
            " for longer text.",
            e.kwargs["message"]
        )

    def test_v2_tags(self):
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "title": "", "subtasks": [],
                                  "tags": {}})
        self.assertEqual(
            "Tags should be an array(list) of strings, but 'dict' is "
            "found.",
            e.kwargs["message"]
        )

        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "title": "", "subtasks": [],
                                  "tags": [1]})
        self.assertEqual(
            "Tag '1' should be a string, but 'int' is found.",
            e.kwargs["message"]
        )

    def test_v2_subtask(self):
        e = self.assertRaises(
            exceptions.InvalidTaskException,
            task_cfg.TaskConfig, {"version": 2, "title": "", "subtasks": {}})
        self.assertEqual(
            "Property 'subtasks' should be an array(list), but 'dict' "
            "is found.",
            e.kwargs["message"]
        )



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""refactor_credentials

Revision ID: 92aaaa2a6bb3
Revises: 4ef544102ba7
Create Date: 2017-02-01 12:52:43.499663

"""
from alembic import op
import sqlalchemy as sa

from rally.common.db import sa_types
from rally import exceptions

# revision identifiers, used by Alembic.
revision = "92aaaa2a6bb3"
down_revision = "4ef544102ba7"
branch_labels = None
depends_on = None


deployments_helper = sa.Table(
    "deployments",
    sa.MetaData(),
    sa.Column("name", sa.String(255), unique=True),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("credentials", sa.PickleType, nullable=True),
    sa.Column("new_credentials", sa_types.MutableJSONEncodedDict,
              default={}, nullable=False)
)


def upgrade():
    with op.batch_alter_table("deployments") as batch_op:
        batch_op.add_column(
            sa.Column("new_credentials", sa_types.MutableJSONEncodedDict,
                      default={}))

    connection = op.get_bind()
    for deployment in connection.execute(deployments_helper.select()):
        creds = {}
        for cred_type, cred_obj in deployment.credentials:
            creds.setdefault(cred_type, [])
            creds[cred_type].append(cred_obj)

        connection.execute(
            deployments_helper.update().where(
                deployments_helper.c.id == deployment.id).values(
                new_credentials=creds))

    with op.batch_alter_table("deployments") as batch_op:
        batch_op.drop_column("credentials")
        batch_op.alter_column("new_credentials", new_column_name="credentials",
                              existing_type=sa_types.MutableJSONEncodedDict,
                              nullable=False)


def downgrade():
    raise exceptions.DowngradeNotSupported()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests

from rally.task import atomic
from rally.task import scenario


class RequestScenario(scenario.Scenario):
    """Base class for Request scenarios with basic atomic actions."""

    @atomic.action_timer("requests.check_request")
    def _check_request(self, url, method, status_code, **kwargs):
        """Compare request status code with specified code

        :param status_code: Expected status code of request
        :param url: Uniform resource locator
        :param method: Type of request method (GET | POST ..)
        :param kwargs: Optional additional request parameters
        :raises ValueError: if return http status code
                            not equal to expected status code
        """

        resp = requests.request(method, url, **kwargs)
        if status_code != resp.status_code:
            error_msg = "Expected HTTP request code is `%s` actual `%s`"
            raise ValueError(
                error_msg % (status_code, resp.status_code))



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io

import ddt
import mock

from rally.cli import cliutils
from rally.cli.commands import plugin as plugin_cmd
from rally.common.plugin import plugin
from rally.common import utils
from tests.unit import test


@ddt.ddt
class PluginCommandsTestCase(test.TestCase):

    def setUp(self):
        super(PluginCommandsTestCase, self).setUp()
        self.plugin_cmd = plugin_cmd.PluginCommands()

        @plugin.configure("p1", "p1_ns")
        class Plugin1(plugin.Plugin):
            """T1.

            Description of T1

            :param x: x arg
            :param y: y arg
            """
            pass

        self.Plugin1 = Plugin1

        @plugin.configure("p2", "p2_ns")
        class Plugin2(plugin.Plugin):
            """T2."""
            pass

        self.Plugin2 = Plugin2

        @plugin.configure("p3", "p2_ns")
        class Plugin3(plugin.Plugin):
            """T3."""
            pass

        self.Plugin3 = Plugin3

    def tearDown(self):
        super(PluginCommandsTestCase, self).tearDown()
        self.Plugin1.unregister()
        self.Plugin2.unregister()
        self.Plugin3.unregister()

    def test__print_plugins_list(self):
        out = io.StringIO()
        original_print_list = cliutils.print_list

        def print_list(*args, **kwargs):
            kwargs["out"] = out
            original_print_list(*args, **kwargs)

        with mock.patch.object(plugin_cmd.cliutils, "print_list",
                               new=print_list):
            plugin_cmd.PluginCommands._print_plugins_list(
                [self.Plugin1, self.Plugin2])

        self.assertEqual(
            "+-------------+------+----------+-------+\n"
            "| Plugin base | Name | Platform | Title |\n"
            "+-------------+------+----------+-------+\n"
            "| Plugin      | p1   | p1_ns    | T1.   |\n"
            "| Plugin      | p2   | p2_ns    | T2.   |\n"
            "+-------------+------+----------+-------+\n", out.getvalue())

    def test_show(self):
        with utils.StdOutCapture() as out:
            plugin_cmd.PluginCommands().show(None, "p1", "p1_ns")
            output = out.getvalue()

            self.assertIn("NAME\n\tp1", output)
            self.assertIn("PLATFORM\n\tp1_ns", output)
            self.assertIn("cli.commands.test_plugin", output)
            self.assertIn("DESCRIPTION\n\tDescription of T1", output)
            self.assertIn("PARAMETERS", output)

    @ddt.data(
        {
            "name": "nonex",
            "platform": None,
            "text": "Plugin nonex not found at any platform\n"
        },
        {
            "name": "nonexplugin",
            "platform": "nonex",
            "text": "Plugin nonexplugin@nonex not found\n"
        }
    )
    @ddt.unpack
    def test_show_not_found(self, name, platform, text):
        with utils.StdOutCapture() as out:
            plugin_cmd.PluginCommands().show(None, name, platform)
            self.assertEqual(out.getvalue(), text)

    @mock.patch("rally.cli.commands.plugin.PluginCommands._print_plugins_list")
    def test_show_many(self, mock_plugin_commands__print_plugins_list):
        with utils.StdOutCapture() as out:
            with mock.patch("rally.cli.commands.plugin.plugin.Plugin."
                            "get_all") as mock_plugin_get_all:
                mock_plugin_get_all.return_value = [self.Plugin2, self.Plugin3]
                plugin_cmd.PluginCommands().show(None, "p", "p2_ns")
                self.assertEqual("Multiple plugins found:\n", out.getvalue())
                mock_plugin_get_all.assert_called_once_with(platform="p2_ns")

        mock_plugin_commands__print_plugins_list.assert_called_once_with([
            self.Plugin2, self.Plugin3])

    @ddt.data(
        {
            "name": None,
            "platform": "nonex",
            "text": "Platform nonex not found\n"
        },
        {
            "name": "p2",
            "platform": "p1_ns",
            "text": "Plugin p2 not found\n"
        }
    )
    @ddt.unpack
    def test_list_not_found(self, name, platform, text):

        with utils.StdOutCapture() as out:
            plugin_cmd.PluginCommands().list(None, name, platform)
            self.assertEqual(text, out.getvalue())

    @mock.patch("rally.cli.commands.plugin.PluginCommands._print_plugins_list")
    def test_list(self, mock_plugin_commands__print_plugins_list):

        plugin_cmd.PluginCommands().list(None, None, "p1_ns")
        plugin_cmd.PluginCommands().list(None, "p1", "p1_ns")
        plugin_cmd.PluginCommands().list(None, "p2", None)

        mock_plugin_commands__print_plugins_list.assert_has_calls([
            mock.call([self.Plugin1]),
            mock.call([self.Plugin1]),
            mock.call([self.Plugin2])
        ])



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
There is a lot of situations when we would like to work with Enum or Const.
E.g. work around Tasks. We would like to use Enum in DB to store status of task
and also in migration that creates DB and in business logic to set some status
so as to avoid copy paste or direct usage of enums values we create singletons
for each enum. (e.g. TaskStatus)
"""

from rally.common import utils


JSON_SCHEMA = "http://json-schema.org/draft-04/schema"
JSON_SCHEMA7 = "http://json-schema.org/draft-07/schema"


class _TaskStatus(utils.ImmutableMixin, utils.EnumMixin):

    """Consts that represents task possible states."""
    INIT = "init"
    VALIDATING = "validating"
    VALIDATED = "validated"
    VALIDATION_FAILED = "validation_failed"
    RUNNING = "running"
    FINISHED = "finished"
    CRASHED = "crashed"
    ABORTING = "aborting"
    SLA_FAILED = "sla_failed"
    SOFT_ABORTING = "soft_aborting"
    ABORTED = "aborted"
    PAUSED = "paused"


class _SubtaskStatus(utils.ImmutableMixin, utils.EnumMixin):

    """Consts that represents task possible states."""
    INIT = "init"
    VALIDATING = "validating"
    VALIDATED = "validated"
    VALIDATION_FAILED = "validation_failed"
    RUNNING = "running"
    FINISHED = "finished"
    CRASHED = "crashed"
    ABORTING = "aborting"
    SLA_FAILED = "sla_failed"
    SOFT_ABORTING = "soft_aborting"
    ABORTED = "aborted"
    PAUSED = "paused"


class _DeployStatus(utils.ImmutableMixin, utils.EnumMixin):
    DEPLOY_INIT = "deploy->init"
    DEPLOY_STARTED = "deploy->started"
    DEPLOY_SUBDEPLOY = "deploy->subdeploy"
    DEPLOY_FINISHED = "deploy->finished"
    DEPLOY_FAILED = "deploy->failed"

    DEPLOY_INCONSISTENT = "deploy->inconsistent"

    CLEANUP_STARTED = "cleanup->started"
    CLEANUP_FINISHED = "cleanup->finished"
    CLEANUP_FAILED = "cleanup->failed"


class _EndpointPermission(utils.ImmutableMixin, utils.EnumMixin):
    ADMIN = "admin"
    USER = "user"


class _EndpointType(utils.ImmutableMixin, utils.EnumMixin):
    INTERNAL = "internal"
    ADMIN = "admin"
    PUBLIC = "public"


class _Service(utils.ImmutableMixin, utils.EnumMixin):
    """OpenStack services names, by rally convention."""

    NOVA = "nova"
    NOVA_NET = "nova-network"
    CINDER = "cinder"
    MANILA = "manila"
    EC2 = "ec2"
    GLANCE = "glance"
    CLOUD = "cloud"
    HEAT = "heat"
    KEYSTONE = "keystone"
    NEUTRON = "neutron"
    DESIGNATE = "designate"
    CEILOMETER = "ceilometer"
    MONASCA = "monasca"
    S3 = "s3"
    SENLIN = "senlin"
    TROVE = "trove"
    SAHARA = "sahara"
    SWIFT = "swift"
    MISTRAL = "mistral"
    MURANO = "murano"
    IRONIC = "ironic"
    GNOCCHI = "gnocchi"
    MAGNUM = "magnum"
    WATCHER = "watcher"


class _ServiceType(utils.ImmutableMixin, utils.EnumMixin):
    """OpenStack services types, mapped to service names."""

    VOLUME = "volume"
    SHARE = "share"
    EC2 = "ec2"
    IMAGE = "image"
    CLOUD = "cloudformation"
    ORCHESTRATION = "orchestration"
    IDENTITY = "identity"
    CLUSTERING = "clustering"
    COMPUTE = "compute"
    NETWORK = "network"
    DNS = "dns"
    METERING = "metering"
    MONITORING = "monitoring"
    S3 = "s3"
    DATABASE = "database"
    DATA_PROCESSING = "data-processing"
    DATA_PROCESSING_MOS = "data_processing"
    OBJECT_STORE = "object-store"
    WORKFLOW_EXECUTION = "workflowv2"
    APPLICATION_CATALOG = "application-catalog"
    BARE_METAL = "baremetal"
    METRIC = "metric"
    CONTAINER_INFRA = "container-infra"
    INFRA_OPTIM = "infra-optim"

    def __init__(self):
        self.__names = {
            self.CLUSTERING: _Service.SENLIN,
            self.COMPUTE: _Service.NOVA,
            self.VOLUME: _Service.CINDER,
            self.SHARE: _Service.MANILA,
            self.EC2: _Service.EC2,
            self.IMAGE: _Service.GLANCE,
            self.CLOUD: _Service.CLOUD,
            self.ORCHESTRATION: _Service.HEAT,
            self.IDENTITY: _Service.KEYSTONE,
            self.NETWORK: _Service.NEUTRON,
            self.DNS: _Service.DESIGNATE,
            self.METERING: _Service.CEILOMETER,
            self.MONITORING: _Service.MONASCA,
            self.S3: _Service.S3,
            self.DATABASE: _Service.TROVE,
            self.DATA_PROCESSING: _Service.SAHARA,
            self.DATA_PROCESSING_MOS: _Service.SAHARA,
            self.OBJECT_STORE: _Service.SWIFT,
            self.WORKFLOW_EXECUTION: _Service.MISTRAL,
            self.APPLICATION_CATALOG: _Service.MURANO,
            self.BARE_METAL: _Service.IRONIC,
            self.METRIC: _Service.GNOCCHI,
            self.CONTAINER_INFRA: _Service.MAGNUM,
            self.INFRA_OPTIM: _Service.WATCHER,
        }

    def __getitem__(self, service_type):
        """Mapping protocol to service names.

        :param name: str, service name
        :returns: str, service type
        """
        return self.__names[service_type]


class _HookStatus(utils.ImmutableMixin, utils.EnumMixin):
    """Hook result statuses."""
    SUCCESS = "success"
    FAILED = "failed"
    VALIDATION_FAILED = "validation_failed"


class _TagType(utils.ImmutableMixin, utils.EnumMixin):
    TASK = "task"
    SUBTASK = "subtask"
    VERIFICATION = "verification"


class _VerifierStatus(utils.ImmutableMixin, utils.EnumMixin):
    """Verifier statuses."""
    INIT = "init"
    INSTALLING = "installing"
    INSTALLED = "installed"
    UPDATING = "updating"
    EXTENDING = "extending"
    FAILED = "failed"


# NOTE(andreykurilin): In case of updating these statuses, please do not forget
#   to update doc reference too
class _VerificationStatus(utils.ImmutableMixin, utils.EnumMixin):
    """Verification statuses."""
    INIT = "init"
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"
    CRASHED = "crashed"


class _TimeFormat(utils.ImmutableMixin, utils.EnumMixin):
    """International time formats"""
    ISO8601 = "%Y-%m-%dT%H:%M:%S%z"


TaskStatus = _TaskStatus()
SubtaskStatus = _SubtaskStatus()
DeployStatus = _DeployStatus()
EndpointPermission = _EndpointPermission()
ServiceType = _ServiceType()
Service = _Service()
EndpointType = _EndpointType()
HookStatus = _HookStatus()
TagType = _TagType()
VerifierStatus = _VerifierStatus()
VerificationStatus = _VerificationStatus()
TimeFormat = _TimeFormat()



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
SQLAlchemy models for rally data.
"""
import datetime as dt
import uuid

import sqlalchemy as sa
import sqlalchemy.ext.declarative
import sqlalchemy.orm   # noqa (used as sa.orm)

from rally.common.db import sa_types
from rally import consts


BASE = sa.ext.declarative.declarative_base()


def UUID():
    return str(uuid.uuid4())


class RallyBase(object):
    """Base class for models."""
    __table_initialized__ = False
    metadata = None

    created_at = sa.Column(sa.DateTime, default=dt.datetime.utcnow)
    updated_at = sa.Column(sa.DateTime, default=dt.datetime.utcnow,
                           onupdate=dt.datetime.utcnow)

    def as_dict(self):
        result = {}
        res = sa.inspect(self)

        for c in self.__table__.columns:
            if c.key not in res.unloaded:
                result[c.name] = getattr(self, c.name)

        for r in self.__mapper__.relationships:
            if r.key not in res.unloaded:
                result[r.key] = getattr(self, r.key)

        return result

    def get(self, key, default=None):
        return getattr(self, key, default)

    def update(self, values):
        """Make the model object behave like a dict."""
        for k, v in values.items():
            setattr(self, k, v)


class Env(BASE, RallyBase):
    """Represent an environment."""
    __tablename__ = "envs"
    __table_args__ = (
        sa.Index("env_uuid", "uuid", unique=True),
        sa.Index("env_name", "name", unique=True),
        sa.Index("env_status", "status")
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text, default="")
    status = sa.Column(sa.String(36), nullable=False)
    extras = sa.Column(sa_types.MutableJSONEncodedDict, default={})
    config = sa.Column(sa_types.MutableJSONEncodedDict, default={})
    spec = sa.Column(sa_types.MutableJSONEncodedDict, default={})


class Platform(BASE, RallyBase):
    """Represent environment's platforms."""
    __tablename__ = "platforms"
    __table_args__ = (
        sa.Index("platform_uuid", "uuid", unique=True),
        sa.Index("platform_env_uuid", "env_uuid")
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)
    env_uuid = sa.Column(sa.String(36), nullable=False)

    status = sa.Column(sa.String(36), nullable=False)

    plugin_name = sa.Column(sa.String(36), nullable=False)
    plugin_spec = sa.Column(sa_types.MutableJSONEncodedDict, default={},
                            nullable=False)
    plugin_data = sa.Column(sa_types.MutableJSONEncodedDict, default={})

    platform_name = sa.Column(sa.String(36))
    platform_data = sa.Column(sa_types.MutableJSONEncodedDict, default={})


class Task(BASE, RallyBase):
    """Represents a task."""
    __tablename__ = "tasks"
    __table_args__ = (
        sa.Index("task_uuid", "uuid", unique=True),
        sa.Index("task_status", "status"),
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)

    env_uuid = sa.Column(sa.String(36), nullable=False)

    # we do not save the whole input task
    input_task = sa.orm.deferred(sa.Column(sa.Text, default=""))

    title = sa.Column(sa.String(128), default="")
    description = sa.Column(sa.Text, default="")

    validation_result = sa.Column(
        sa_types.MutableJSONEncodedDict, default={}, nullable=False)

    # we do not calculate the duration of a validation step yet
    validation_duration = sa.orm.deferred(sa.Column(sa.Float))

    task_duration = sa.Column(sa.Float, default=0.0)
    pass_sla = sa.Column(sa.Boolean, default=True)
    status = sa.Column(sa.String(36), default=consts.TaskStatus.INIT)


class Subtask(BASE, RallyBase):
    __tablename__ = "subtasks"
    __table_args__ = (
        sa.Index("subtask_uuid", "uuid", unique=True),
        sa.Index("subtask_status", "status"),
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)

    task_uuid = sa.Column(
        sa.String(36),
        sa.ForeignKey(Task.uuid),
        nullable=False,
    )

    task = sa.orm.relationship(
        Task,
        backref=sa.orm.backref("subtasks"),
        foreign_keys=task_uuid,
        primaryjoin=(task_uuid == Task.uuid),
    )

    title = sa.Column(sa.String(128), default="")
    description = sa.Column(sa.Text, default="")

    # we do not support subtask contexts feature yet, see
    # https://review.openstack.org/#/c/404168/
    contexts = sa.orm.deferred(sa.Column(
        sa_types.JSONEncodedDict, default={}, nullable=False))

    contexts_results = sa.orm.deferred(sa.Column(
        sa_types.MutableJSONEncodedList, default=[], nullable=False))

    sla = sa.Column(
        sa_types.JSONEncodedDict, default={}, nullable=False)

    # It is always False for now
    run_in_parallel = sa.orm.deferred(
        sa.Column(sa.Boolean, default=False, nullable=False))

    duration = sa.Column(sa.Float, default=0.0)
    pass_sla = sa.Column(sa.Boolean, default=True)
    status = sa.Column(sa.String(36), default=consts.SubtaskStatus.RUNNING)


class Workload(BASE, RallyBase):
    __tablename__ = "workloads"
    __table_args__ = (
        sa.Index("workload_uuid", "uuid", unique=True),
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)

    task_uuid = sa.Column(
        sa.String(36),
        sa.ForeignKey(Task.uuid),
        nullable=False,
    )

    subtask_uuid = sa.Column(
        sa.String(36),
        sa.ForeignKey(Subtask.uuid),
        nullable=False,
    )

    subtask = sa.orm.relationship(
        Subtask,
        backref=sa.orm.backref("workloads"),
        foreign_keys=subtask_uuid,
        primaryjoin=(subtask_uuid == Subtask.uuid),
    )

    name = sa.Column(sa.String(64), nullable=False)
    description = sa.Column(sa.Text, default="")
    position = sa.Column(sa.Integer, default=0, nullable=False)

    runner = sa.Column(
        sa_types.JSONEncodedDict, default={}, nullable=False)

    runner_type = sa.Column(sa.String(64), nullable=False)

    contexts = sa.Column(
        sa_types.JSONEncodedDict, default={}, nullable=False)

    contexts_results = sa.Column(
        sa_types.MutableJSONEncodedList, default=[], nullable=False)

    sla = sa.Column(
        sa_types.JSONEncodedDict, default={}, nullable=False)

    sla_results = sa.Column(
        sa_types.MutableJSONEncodedDict, default={}, nullable=False)

    args = sa.Column(
        sa_types.JSONEncodedDict, default={}, nullable=False)

    hooks = sa.Column(
        sa_types.JSONEncodedList, default=[], nullable=False)

    start_time = sa.Column(sa_types.TimeStamp)

    load_duration = sa.Column(sa.Float, default=0.0)
    full_duration = sa.Column(sa.Float, default=0.0)
    min_duration = sa.Column(sa.Float)
    max_duration = sa.Column(sa.Float)
    total_iteration_count = sa.Column(sa.Integer, default=0)
    failed_iteration_count = sa.Column(sa.Integer, default=0)

    statistics = sa.Column(
        sa_types.MutableJSONEncodedDict, default={}, nullable=False)

    pass_sla = sa.Column(sa.Boolean, default=True)
    _profiling_data = sa.orm.deferred(sa.Column(sa.Text, default=""))


class WorkloadData(BASE, RallyBase):
    __tablename__ = "workloaddata"
    __table_args__ = (
        sa.Index("workload_data_uuid", "uuid", unique=True),
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)

    task_uuid = sa.Column(
        sa.String(36),
        sa.ForeignKey(Task.uuid),
        nullable=False,
    )

    workload_uuid = sa.Column(
        sa.String(36),
        sa.ForeignKey(Workload.uuid),
        nullable=False,
    )

    workload = sa.orm.relationship(
        Workload,
        backref=sa.orm.backref("workload_data"),
        foreign_keys=workload_uuid,
        primaryjoin=(workload_uuid == Workload.uuid),
    )

    chunk_order = sa.Column(sa.Integer, nullable=False)
    chunk_data = sa.Column(
        sa_types.MutableJSONEncodedDict, default={}, nullable=False)
    # all these fields are not used
    iteration_count = sa.orm.deferred(sa.Column(sa.Integer, nullable=False))
    failed_iteration_count = sa.orm.deferred(sa.Column(
        sa.Integer, nullable=False))
    chunk_size = sa.orm.deferred(sa.Column(
        sa.Integer, nullable=False))
    compressed_chunk_size = sa.orm.deferred(sa.Column(
        sa.Integer, nullable=False))
    started_at = sa.orm.deferred(sa.Column(
        sa.DateTime, default=dt.datetime.utcnow, nullable=False))
    finished_at = sa.orm.deferred(sa.Column(
        sa.DateTime, default=dt.datetime.utcnow, nullable=False))


class Tag(BASE, RallyBase):
    __tablename__ = "tags"
    __table_args__ = (
        sa.Index("d_type_tag", "uuid", "type", "tag", unique=True),
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)
    type = sa.Column(sa.String(36), nullable=False)
    tag = sa.Column(sa.String(255), nullable=False)


class Verifier(BASE, RallyBase):
    """Represents a verifier."""

    __tablename__ = "verifiers"
    __table_args__ = (
        sa.Index("verifier_uuid", "uuid", unique=True),
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)

    name = sa.Column(sa.String(255), unique=True)
    description = sa.Column(sa.Text)

    type = sa.Column(sa.String(255), nullable=False)
    platform = sa.Column(sa.String(255))

    source = sa.Column(sa.String(255))
    version = sa.Column(sa.String(255))
    system_wide = sa.Column(sa.Boolean)

    status = sa.Column(sa.String(36), default=consts.VerifierStatus.INIT,
                       nullable=False)

    extra_settings = sa.Column(sa_types.MutableJSONEncodedDict)


class Verification(BASE, RallyBase):
    """Represents a verification."""

    __tablename__ = "verifications"
    __table_args__ = (
        sa.Index("verification_uuid", "uuid", unique=True),
    )

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(sa.String(36), default=UUID, nullable=False)

    verifier_uuid = sa.Column(sa.String(36),
                              sa.ForeignKey(Verifier.uuid),
                              nullable=False)
    env_uuid = sa.Column(sa.String(36), nullable=False)

    run_args = sa.Column(sa_types.MutableJSONEncodedDict)

    status = sa.Column(sa.String(36), default=consts.VerificationStatus.INIT,
                       nullable=False)

    tests_count = sa.Column(sa.Integer, default=0)
    failures = sa.Column(sa.Integer, default=0)
    skipped = sa.Column(sa.Integer, default=0)
    success = sa.Column(sa.Integer, default=0)
    unexpected_success = sa.Column(sa.Integer, default=0)
    expected_failures = sa.Column(sa.Integer, default=0)
    tests_duration = sa.Column(sa.Float, default=0.0)

    tests = sa.Column(sa_types.MutableJSONEncodedDict, default={})



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import decorator

from rally.common.plugin import discover


PLUGINS_LOADED = False


def load():
    global PLUGINS_LOADED

    if not PLUGINS_LOADED:
        from rally.common import opts

        opts.register()

        discover.import_modules_from_package("rally.plugins.common")

        packages = discover.find_packages_by_entry_point()
        for package in packages:
            if "options" in package:
                opts.register_options_from_path(package["options"])
        discover.import_modules_by_entry_point(_packages=packages)

        discover.load_plugins("/opt/rally/plugins/")
        discover.load_plugins(os.path.expanduser("~/.rally/plugins/"))

    PLUGINS_LOADED = True


@decorator.decorator
def ensure_plugins_are_loaded(f, *args, **kwargs):
    load()
    return f(*args, **kwargs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.verification import context
from tests.unit import test


@context.configure("fake_verifier_context", order=314)
class FakeContext(context.VerifierContext):
    def cleanup(self):
        pass

    def setup(self):
        super(FakeContext, self).setup()
        pass


class VerifierContextTestCase(test.TestCase):
    def test__meta_get(self):

        data = {"key1": "value1", "key2": "value2"}

        for k, v in data.items():
            FakeContext._meta_set(k, v)

        for k, v in data.items():
            self.assertEqual(v, FakeContext._meta_get(k))

        self.assertTrue(FakeContext.is_hidden())


class ContextManagerTestCase(test.TestCase):
    @mock.patch("rally.verification.context.VerifierContext")
    def test_validate(self, mock_verifier_context):
        config = {"ctx1": mock.Mock(), "ctx2": mock.Mock()}

        context.ContextManager.validate(config)

        self.assertEqual([mock.call(k, allow_hidden=True)
                          for k, v in config.items()],
                         mock_verifier_context.get.call_args_list)
        self.assertEqual(
            [mock.call(v) for k, v in config.items()],
            mock_verifier_context.get.return_value.validate.call_args_list)



#!/usr/bin/env python
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import collections
import gzip
import json
import logging
import os
import re
import subprocess
import sys
import uuid

from rally_openstack import osclients

from rally import api
from rally.ui import utils

LOG = logging.getLogger("verify-job")
LOG.setLevel(logging.DEBUG)

# NOTE(andreykurilin): this variable is used to generate output file names
# with prefix ${CALL_COUNT}_ .
_call_count = 0


class Status(object):
    PASS = "success"
    ERROR = "error"
    SKIPPED = "skip"
    FAILURE = "fail"


class Step(object):
    COMMAND = None
    DEPENDS_ON = None
    CALL_ARGS = {}

    BASE_DIR = "rally-verify"
    HTML_TEMPLATE = ("<span class=\"%(status)s\">[%(status)s]</span>\n"
                     "<a href=\"%(output_file)s\">%(doc)s</a>\n"
                     "<code>$ %(cmd)s</code>")

    def __init__(self, args, rapi):
        self.args = args
        self.rapi = rapi
        self.result = {"status": Status.PASS,
                       "doc": self.__doc__}

    @property
    def name(self):
        return " ".join(re.findall("[A-Z][^A-Z]*",
                                   self.__class__.__name__)).lower()

    def check(self, results):
        """Check weather this step should be executed or skipped."""
        if self.DEPENDS_ON is not None:
            if results[self.DEPENDS_ON].result["status"] in (
                    Status.PASS, Status.FAILURE):
                return True
            else:
                self.result["status"] = Status.SKIPPED
                msg = ("Step '%s' is skipped, since depends on step '%s' is "
                       "skipped or finished with an error." %
                       (self.name, results[self.DEPENDS_ON].name))
                stdout_file = self._generate_path(
                    "%s.txt.gz" % self.__class__.__name__)

                self.result["output_file"] = self._write_file(
                    stdout_file, msg, compress=True)
                return False
        return True

    def setUp(self):
        """Obtain variables required for execution"""
        pass

    def run(self):
        """Execute step. The default action - execute the command"""
        self.setUp()

        cmd = "rally --rally-debug %s" % (self.COMMAND % self.CALL_ARGS)
        self.result["cmd"] = cmd
        self.result["status"], self.result["output"] = self.call_rally(cmd)

        stdout_file = self._generate_path("%s.txt.gz" % cmd)
        self.result["output_file"] = self._write_file(
            stdout_file, self.result["output"], compress=True)

    @classmethod
    def _generate_path(cls, root):
        global _call_count
        _call_count += 1

        root = root.replace("<", "").replace(">", "").replace("/", "_")
        parts = ["%s" % _call_count]
        for path in root.split(" "):
            if path.startswith(cls.BASE_DIR):
                path = path[len(cls.BASE_DIR) + 1:]
            parts.append(path)
        return os.path.join(cls.BASE_DIR, "_".join(parts))

    @classmethod
    def _write_file(cls, path, data, compress=False):
        """Create a file and write some data to it."""
        if compress:
            with gzip.open(path, "wb") as f:
                f.write(data)
        else:
            with open(path, "wb") as f:
                f.write(data)
        return path

    @staticmethod
    def call_rally(command):
        """Execute a Rally verify command."""
        try:
            LOG.info("Start `%s` command." % command)
            stdout = subprocess.check_output(command.split(),
                                             stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            LOG.error("Command `%s` failed." % command)
            return Status.ERROR, e.output
        else:
            return Status.PASS, stdout

    def to_html(self):
        return self.HTML_TEMPLATE % self.result


class SetUpStep(Step):
    """Validate deployment, create required resources and directories."""

    DEPLOYMENT_NAME = "devstack"

    def run(self):
        if not os.path.exists("%s/extra" % self.BASE_DIR):
            os.makedirs("%s/extra" % self.BASE_DIR)

        # ensure that deployment exit
        deployment = self.rapi.deployment._get(self.DEPLOYMENT_NAME)
        # check it
        result = self.rapi.deployment.check(
            deployment=self.DEPLOYMENT_NAME)["openstack"]
        if "admin_error" in result[0] or "user_error" in result[0]:
            self.result["status"] = Status.ERROR
            return

        try:
            subprocess.check_call(["rally", "deployment", "use",
                                   "--deployment", self.DEPLOYMENT_NAME],
                                  stdout=sys.stdout)
        except subprocess.CalledProcessError:
            self.result["status"] = Status.ERROR
            return

        credentials = None
        for platform, creds in deployment.to_dict()["credentials"].items():
            if platform == "openstack":
                credentials = creds[0]["admin"]
        if credentials is None:
            return Status.ERROR, "There is no openstack credentials."

        clients = osclients.Clients(credentials)

        if self.args.ctx_create_resources:
            # If the 'ctx-create-resources' arg is provided, delete images and
            # flavors, and also create a shared network to make Tempest context
            # create needed resources.
            LOG.info("The 'ctx-create-resources' arg is provided. Deleting "
                     "images and flavors, and also creating a shared network "
                     "to make Tempest context create needed resources.")

            LOG.info("Deleting images.")
            for image in clients.glance().images.list():
                clients.glance().images.delete(image.id)

            LOG.info("Deleting flavors.")
            for flavor in clients.nova().flavors.list():
                clients.nova().flavors.delete(flavor.id)

            LOG.info("Creating a shared network.")
            net_body = {
                "network": {
                    "name": "shared-net-%s" % str(uuid.uuid4()),
                    "tenant_id": clients.keystone.auth_ref.project_id,
                    "shared": True
                }
            }
            clients.neutron().create_network(net_body)
        else:
            # Otherwise, just in case create only flavors with the following
            # properties: RAM = 64MB and 128MB, VCPUs = 1, disk = 0GB to make
            # Tempest context discover them.
            LOG.info("The 'ctx-create-resources' arg is not provided. "
                     "Creating flavors to make Tempest context discover them.")
            for flv_ram in [64, 128]:
                params = {
                    "name": "flavor-%s" % str(uuid.uuid4()),
                    "ram": flv_ram,
                    "vcpus": 1,
                    "disk": 0
                }
                LOG.info("Creating flavor '%s' with the following properties: "
                         "RAM = %dMB, VCPUs = 1, disk = 0GB" %
                         (params["name"], flv_ram))
                clients.nova().flavors.create(**params)

    def to_html(self):
        return ""


class ListPlugins(Step):
    """List plugins for verifiers management."""

    COMMAND = "verify list-plugins"
    DEPENDS_ON = SetUpStep


class CreateVerifier(Step):
    """Create a Tempest verifier."""

    COMMAND = ("verify create-verifier --type %(type)s --name %(name)s "
               "--source %(source)s")
    DEPENDS_ON = ListPlugins
    CALL_ARGS = {"type": "tempest",
                 "name": "my-verifier",
                 "source": "https://git.openstack.org/openstack/tempest"}


class ShowVerifier(Step):
    """Show information about the created verifier."""

    COMMAND = "verify show-verifier"
    DEPENDS_ON = CreateVerifier


class ListVerifiers(Step):
    """List all installed verifiers."""

    COMMAND = "verify list-verifiers"
    DEPENDS_ON = CreateVerifier


class UpdateVerifier(Step):
    """Switch the verifier to the penultimate version."""

    COMMAND = "verify update-verifier --version %(version)s --update-venv"
    DEPENDS_ON = CreateVerifier

    def setUp(self):
        """Obtain penultimate verifier commit for downgrading to it"""
        verifier_id = self.rapi.verifier.list()[0]["uuid"]
        verifications_dir = os.path.join(
            os.path.expanduser("~"),
            ".rally/verification/verifier-%s/repo" % verifier_id)
        # Get the penultimate verifier commit ID
        p_commit_id = subprocess.check_output(
            ["git", "log", "-n", "1", "--pretty=format:%H"],
            cwd=verifications_dir).strip()
        self.CALL_ARGS = {"version": p_commit_id}


class ConfigureVerifier(Step):
    """Generate and show the verifier config file."""

    COMMAND = "verify configure-verifier --show"
    DEPENDS_ON = CreateVerifier


class ExtendVerifier(Step):
    """Extend verifier with keystone integration tests."""

    COMMAND = "verify add-verifier-ext --source %(source)s"
    DEPENDS_ON = CreateVerifier
    CALL_ARGS = {"source": "https://git.openstack.org/openstack/"
                           "keystone-tempest-plugin"}


class ListVerifierExtensions(Step):
    """List all extensions of verifier."""

    COMMAND = "verify list-verifier-exts"
    DEPENDS_ON = ExtendVerifier


class ListVerifierTests(Step):
    """List all tests of specific verifier."""

    COMMAND = "verify list-verifier-tests"
    DEPENDS_ON = CreateVerifier


class RunVerification(Step):
    """Run a verification."""

    DEPENDS_ON = ConfigureVerifier
    COMMAND = ("verify start --pattern set=%(set)s --skip-list %(skip_tests)s "
               "--xfail-list %(xfail_tests)s --tag %(tag)s %(set)s-set "
               "--detailed")
    SKIP_TESTS = {
        "tempest.api.compute.flavors.test_flavors.FlavorsV2TestJSON."
        "test_get_flavor[id-1f12046b-753d-40d2-abb6-d8eb8b30cb2f,smoke]":
            "This test was skipped intentionally"}
    XFAIL_TESTS = {
        "tempest.api.compute.servers.test_server_actions."
        "ServerActionsTestJSON.test_get_vnc_console"
        "[id-c6bc11bf-592e-4015-9319-1c98dc64daf5]":
            "This test fails because 'novnc' console type is unavailable"}

    def setUp(self):
        self.CALL_ARGS["tag"] = "tag-1 tag-2"
        self.CALL_ARGS["set"] = "full" if self.args.mode == "full" else "smoke"
        # Start a verification, show results and generate reports
        skip_tests = json.dumps(self.SKIP_TESTS)
        xfail_tests = json.dumps(self.XFAIL_TESTS)
        self.CALL_ARGS["skip_tests"] = self._write_file(
            self._generate_path("skip-list.json"), skip_tests)
        self.CALL_ARGS["xfail_tests"] = self._write_file(
            self._generate_path("xfail-list.json"), xfail_tests)

    def run(self):
        super(RunVerification, self).run()
        if "Success: 0" in self.result["output"]:
            self.result["status"] = Status.FAILURE


class ReRunVerification(RunVerification):
    """Re-Run previous verification."""

    COMMAND = "verify rerun --tag one-more-attempt"


class ShowVerification(Step):
    """Show results of verification."""

    COMMAND = "verify show"
    DEPENDS_ON = RunVerification


class ShowSecondVerification(ShowVerification):
    """Show results of verification."""

    DEPENDS_ON = ReRunVerification


class ShowDetailedVerification(Step):
    """Show detailed results of verification."""

    COMMAND = "verify show --detailed"
    DEPENDS_ON = RunVerification


class ShowDetailedSecondVerification(ShowDetailedVerification):
    """Show detailed results of verification."""

    DEPENDS_ON = ReRunVerification


class ReportVerificationMixin(Step):
    """Mixin for obtaining reports of verifications."""

    COMMAND = "verify report --uuid %(uuids)s --type %(type)s --to %(out)s"

    HTML_TEMPLATE = ("<span class=\"%(status)s\">[%(status)s]</span>\n"
                     "<a href=\"%(out)s\">%(doc)s</a> "
                     "[<a href=\"%(output_file)s\">Output from CLI</a>]\n"
                     "<code>$ %(cmd)s</code>")

    def setUp(self):
        self.CALL_ARGS["out"] = "<path>"
        self.CALL_ARGS["uuids"] = "<uuid-1> <uuid-2>"
        cmd = self.COMMAND % self.CALL_ARGS
        report = "%s.%s" % (cmd.replace("/", "_").replace(" ", "_"),
                            self.CALL_ARGS["type"])
        print(report)
        self.CALL_ARGS["out"] = self._generate_path(report)
        self.CALL_ARGS["uuids"] = " ".join(
            [v["uuid"] for v in self.rapi.verification.list()])
        print(self.COMMAND % self.CALL_ARGS)

    def run(self):
        super(ReportVerificationMixin, self).run()
        creport = "%s.gz" % self.CALL_ARGS["out"]
        with open(self.CALL_ARGS["out"], "rb") as f_in:
            with gzip.open(creport, "wb") as f_out:
                f_out.writelines(f_in)
        self.result["out"] = creport


class HtmlVerificationReport(ReportVerificationMixin):
    """Generate HTML report for verification(s)."""

    CALL_ARGS = {"type": "html-static"}
    DEPENDS_ON = RunVerification

    def setUp(self):
        super(HtmlVerificationReport, self).setUp()
        self.CALL_ARGS["out"] = self.CALL_ARGS["out"][:-7]


class JsonVerificationReport(ReportVerificationMixin):
    """Generate JSON report for verification(s)."""

    CALL_ARGS = {"type": "json"}
    DEPENDS_ON = RunVerification


class JunitVerificationReport(ReportVerificationMixin):
    """Generate JUNIT report for verification(s)."""

    CALL_ARGS = {"type": "junit-xml"}
    DEPENDS_ON = RunVerification


class ListVerifications(Step):
    """List all verifications."""

    COMMAND = "verify list"
    DEPENDS_ON = CreateVerifier


class DeleteVerifierExtension(Step):
    """Delete keystone extension."""

    COMMAND = "verify delete-verifier-ext --name %(name)s"
    CALL_ARGS = {"name": "keystone_tests"}
    DEPENDS_ON = ExtendVerifier


class DeleteVerifier(Step):
    """Delete only Tempest verifier.

    all verifications will be delete when destroy deployment.

    """
    COMMAND = "verify delete-verifier --id %(id)s --force"
    CALL_ARGS = {"id": CreateVerifier.CALL_ARGS["name"]}
    DEPENDS_ON = CreateVerifier


class DestroyDeployment(Step):
    """Delete the deployment, and verifications of this deployment."""

    COMMAND = "deployment destroy --deployment %(id)s"
    CALL_ARGS = {"id": SetUpStep.DEPLOYMENT_NAME}
    DEPENDS_ON = SetUpStep


def run(args):

    steps = [SetUpStep,
             ListPlugins,
             CreateVerifier,
             ShowVerifier,
             ListVerifiers,
             UpdateVerifier,
             ConfigureVerifier,
             ExtendVerifier,
             ListVerifierExtensions,
             ListVerifierTests,
             RunVerification,
             ShowVerification,
             ShowDetailedVerification,
             HtmlVerificationReport,
             JsonVerificationReport,
             JunitVerificationReport,
             ListVerifications,
             DeleteVerifierExtension,
             DestroyDeployment,
             DeleteVerifier]

    if args.compare:
        # need to launch one more verification
        place_to_insert = steps.index(ShowDetailedVerification) + 1
        # insert steps in reverse order to be able to use the same index
        steps.insert(place_to_insert, ShowDetailedSecondVerification)
        steps.insert(place_to_insert, ShowSecondVerification)
        steps.insert(place_to_insert, ReRunVerification)

    results = collections.OrderedDict()
    rapi = api.API()
    for step_cls in steps:
        step = step_cls(args, rapi=rapi)
        if step.check(results):
            step.run()
        results[step_cls] = step

    return results.values()


def main():
    parser = argparse.ArgumentParser(description="Launch rally-verify job.")
    parser.add_argument("--mode", type=str, default="light",
                        help="Mode of job. The 'full' mode corresponds to the "
                             "full set of verifier tests. The 'light' mode "
                             "corresponds to the smoke set of verifier tests.",
                        choices=["light", "full"])
    parser.add_argument("--compare", action="store_true",
                        help="Start the second verification to generate a "
                             "trends report for two verifications.")
    # TODO(ylobankov): Remove hard-coded Tempest related things and make it
    #                  configurable.
    parser.add_argument("--ctx-create-resources", action="store_true",
                        help="Make Tempest context create needed resources "
                             "for the tests.")
    return

    args = parser.parse_args()

    steps = run(args)
    results = [step.to_html() for step in steps]

    template = utils.get_template("ci/index_verify.html")
    with open(os.path.join(Step.BASE_DIR, "extra/index.html"), "w") as f:
        f.write(template.render(steps=results))

    if len([None for step in steps
            if step.result["status"] == Status.PASS]) == len(steps):
        return 0
    return 1


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

from rally.plugins.common.hook.triggers import periodic
from rally.task import hook
from tests.unit import test


@ddt.ddt
class PeriodicTriggerTestCase(test.TestCase):

    def setUp(self):
        super(PeriodicTriggerTestCase, self).setUp()
        self.hook_cls = mock.MagicMock(__name__="name")
        self.trigger = periodic.PeriodicTrigger(
            {"trigger": ("periodic", {"unit": "iteration", "step": 2}),
             "action": ("foo", {})},
            mock.MagicMock(), self.hook_cls)

    @ddt.data((dict(unit="time", step=1), True),
              (dict(unit="time", step=0), False),
              (dict(unit="time", step=1, start=0), True),
              (dict(unit="time", step=1, start=-1), False),
              (dict(unit="time", step=1, start=0, end=1), True),
              (dict(unit="time", step=1, start=0, end=0), False),
              (dict(unit="time", wrong_prop=None), False),
              (dict(unit="time"), False),
              (dict(unit="iteration", step=1), True),
              (dict(unit="iteration", step=0), False),
              (dict(unit="iteration", step=1, start=1), True),
              (dict(unit="iteration", step=1, start=0), False),
              (dict(unit="iteration", step=1, start=1, end=1), True),
              (dict(unit="iteration", step=1, start=1, end=0), False),
              (dict(unit="iteration", wrong_prop=None), False),
              (dict(unit="iteration"), False),
              (dict(unit="wrong-unit", step=1), False),
              (dict(step=1), False))
    @ddt.unpack
    def test_validate(self, config, valid):
        results = hook.HookTrigger.validate("periodic", None, None, config)
        if valid:
            self.assertEqual([], results)
        else:
            self.assertEqual(1, len(results))

    def test_get_listening_event(self):
        event_type = self.trigger.get_listening_event()
        self.assertEqual("iteration", event_type)

    @ddt.data((1, True), (2, False), (3, True), (4, False), (5, True),
              (6, False), (7, True), (8, False), (9, True), (10, False))
    @ddt.unpack
    def test_on_event(self, value, should_call):
        self.trigger.on_event("iteration", value)
        self.assertEqual(should_call, self.hook_cls.called)

    @ddt.data((0, False), (1, False), (2, True), (3, False), (4, False),
              (5, True), (6, False), (7, False), (8, True), (9, False))
    @ddt.unpack
    def test_on_event_start_end(self, value, should_call):
        trigger = periodic.PeriodicTrigger(
            {"trigger": ("periodic", {"unit": "time",
                                      "step": 3, "start": 2, "end": 9}),
             "action": ("foo", {})},
            mock.MagicMock(), self.hook_cls)
        trigger.on_event("time", value)
        self.assertEqual(should_call, self.hook_cls.called)
